from unittest import TestCase
import pandas as pd
from landbosse.model import ErectionCost
import os
from landbosse.excelio import XlsxReader
from landbosse.tests.model.test_filename_functions import landbosse_test_input_dir
import logging
import sys


log = logging.getLogger(__name__)
out_hdlr = logging.StreamHandler(sys.stdout)
# out_hdlr.setFormatter(logging.Formatter('%(asctime)s %(message)s'))
out_hdlr.setFormatter(logging.Formatter('%(message)s'))
out_hdlr.setLevel(logging.DEBUG)
log.addHandler(out_hdlr)
log.setLevel(logging.DEBUG)

class TestErectionCost(TestCase):
    def setUp(self):
        print('<><>><><><><><><><><> Begin load of ErectionCost test data <><>><><><><><><><><>')

        self.erection_cost_input_dict = None
        project = pd.Series(
            {
                'Turbine rating MW': 1.5,
                'Hub height m': 80,
                'Rotor diameter m': 77,
                'Number of breakdowns per 100 MW': 4,
                'Number of sections per tower': 3,
                'Number of RNA picks (at hub height)': 4,
                'Number of turbines': 100,
                'Breakpoint between base and topping (percent)': 0.75,
                'Fuel cost USD per gal': 1,
                'Rate of deliveries (turbines per week)': 10,
                'Wind shear exponent': 0.2,
                'Tower type': 'steel',
                'Foundation depth m': 3,
                'Rated Thrust (N)': 100,
                'Bearing Pressure (n/m2)': 100,
                'Critical Velocity (m/s)': 20,
                'Line Frequency (Hz)': 60,
                'Project ID': 'foo',
                'Turbine spacing (times rotor diameter)': 5,
                'Row spacing (times rotor diameter)': 5,
                'Number of turbines per day': 2,
                'Flag for user-defined home run trench length (0 = no; 1 = yes)': 1,
                'Combined Homerun Trench Length to Substation (km)': 67,
                'Total project construction time (months)': 9,
                '50-year Gust Velocity (m/s)': 60,
                'Road length adder (m)': 5000,
                'Percent of roads that will be constructed': 0.33,
                'Road Quality (0-1)': 0.6,
                'Line Frequency (Hz)': 60,
                'Row spacing (times rotor diameter)': 10,
                'Flag for user-defined home run trench length (0 = no; 1 = yes)': 0,
                'Combined Homerun Trench Length to Substation (km)': 50,
                'Distance to interconnect (miles)': 50,
                'Interconnect Voltage (kV)': 137,
                'New Switchyard (y/n)': 'y',
                'Non-Erection Wind Delay Critical Speed (m/s)': 15,
                'Non-Erection Wind Delay Critical Height (m)': 10,
                'Road width (ft)': 20,
                'Road thickness (in)': 8,
                'Crane width (m)': 12.2,
                'Number of highway permits': 10,
                'Number of access roads': 2,
                'Overtime multiplier': 1.4,
                'Allow same flag': 'n',
                'Markup contingency': 0.03,
                'Markup warranty management': 0.0002,
                'Markup sales and use tax': 0,
                'Markup overhead': 0.05,
                'Markup profit margin': 0.05
            }
        )

        xlsx_reader = XlsxReader()
        input_xlsx = os.path.join(landbosse_test_input_dir(), 'erection_cost_tests.xlsx')
        self.master_input_dict = xlsx_reader.create_master_input_dictionary(input_xlsx=input_xlsx, project_parameters=project)

        print('<><>><><><><><><><><> End load of ErectionCost test data <><>><><><><><><><><>')

    def key_value_logging_helper(self, _dict):
        for key, value in _dict.items():
            if isinstance(value, dict):
                print(f'{key}:')
                self.key_value_logging_helper(value)
            else:
                print(key, ':\n' if isinstance(value, pd.DataFrame) else ':', value)

    def test_ErectionCostModule_Black_Box(self):
        print('>>>>>>>>>>>>>>>>>>>>> Begin ErectionCost Module black box test <<<<<<<<<<<<<<<<<<<')
        print('===================== Inputs =====================')
        self.key_value_logging_helper(self.master_input_dict)
        print('===================== Outputs =====================')
        erection_cost_output_dict = dict()

        ec = ErectionCost(input_dict=self.master_input_dict, output_dict=erection_cost_output_dict, log=log, project_name='foo')
        ec.run_module()
        self.key_value_logging_helper(erection_cost_output_dict)
        print('>>>>>>>>>>>>>>>>>>>>> End ErectionCost Module black box test <<<<<<<<<<<<<<<<<<<')
        self.assertTrue(True)



import pandas as pd


class CsvGenerator:
    """
    This class generates CSV files.
    """

    def __init__(self, file_ops):
        """
        Parameters
        ----------
        file_ops : XlsxFileOperations
            An instance of XlsxFileOperations to manage file names.
        """
        self.file_ops = file_ops

    def create_details_dataframe(self, details):
        """
        This writes the details .csv.

        Parameters
        ----------
        details : list[dict]
            A list of dictionaries to be converted into a Pandas dataframe

        Returns
        -------
        pd.DataFrame
            The dataframe that can be written to a .csv file.
        """

        # This the list of details to write to the .csv
        details_to_write_to_csv = []
        for row in details:
            new_row = {}
            new_row["Project ID with serial"] = row["project_id_with_serial"]
            new_row["Module"] = row["module"]
            new_row["Variable name"] = row["variable_df_key_col_name"]
            new_row["Unit"] = row["unit"]

            value = row["value"]
            value_is_number = self._is_numeric(value)
            if value_is_number:
                new_row["Numeric value"] = value
            else:
                new_row["Non-numeric value"] = value

            # If there is a last_number, which means this is a dataframe row that has a number
            # at the end, write this into the numeric value column. This overrides automatic
            # type detection.

            if "last_number" in row:
                new_row["Numeric value"] = row["last_number"]

            details_to_write_to_csv.append(new_row)

        details = pd.DataFrame(details_to_write_to_csv)

        return details

    def create_costs_dataframe(self, costs):
        """
        Parameters
        ----------
        costs : list[dict]
            The list of dictionaries of costs.

        Returns
        -------
        pd.DataFrame
            A dataframe to be written as a .csv
        """
        new_rows = []
        for row in costs:
            new_row = {
                "Project ID with serial": row["project_id_with_serial"],
                "Number of turbines": row["num_turbines"],
                "Turbine rating MW": row["turbine_rating_MW"],
                "Rotor diameter m": row["rotor_diameter_m"],
                "Module": row["module"],
                "Type of cost": row["type_of_cost"],
                "Cost per turbine": row["cost_per_turbine"],
                "Cost per project": row["cost_per_project"],
                "Cost per kW": row["usd_per_kw_per_project"]
            }
            new_rows.append(new_row)
        costs_df = pd.DataFrame(new_rows)
        return costs_df

    def _is_numeric(self, value):
        """
        This method tests if a value is a numeric (that is, can be parsed
        by float()) or non numeric (which cannot be parsed).

        The decision from this method determines whether values go into
        the numeric or non-numeric columns.

        Parameters
        ----------
        value
            The value to be tested.

        Returns
        -------
        bool
            True if the value is numeric, False otherwise.
        """
        try:
            float(value)
        except ValueError:
            return False
        return True



import pandas as pd

from .XlsxDataframeCache import XlsxDataframeCache
from .XlsxFileOperations import XlsxFileOperations
from .XlsxReader import XlsxReader


class XlsxManagerRunner:
    """
    This class runs a new instance of Manager for each project listed
    in the project_xlsx file.

    This class is meant to be subclassed depending on whether a serial
    or parallel manager runner is needed.
    """

    def __init__(self, file_ops=None):
        """
        The constructor simply creates an XlsxFileOperations instance
        to live throughout the lifetime of the instance

        Parameters
        ----------
        file_ops : XlsxFileOperations
            The file operation instance used to create filenames. If this
            is left at the default of None, a new instance of
            XlsxFileOperations is created.
        """
        self.file_ops = file_ops if file_ops is not None else XlsxFileOperations()

    def run_from_project_list_xlsx(self, projects_xlsx,  enable_cost_and_scaling_modifications=True):
        """
        This function runs all the scenarios in the projects_xlsx file. It creates
        the OrderedDict that holds the results of all the runs. See the return
        section below for more details on what the OrderedDict contains.

        This method is meant to be overriden by subclasses. If this method is
        called directly on this class, a NotImplementedError is raised. However, this
        docstring does specify the contract that the inheriting class should raise.

        Parameters
        ----------
        projects_xlsx : str
            A path name (preferably created with os.path.join()) specific to the
            operating system that is the main input .xlsx file that controls
            running of all the projects. Crucially, this file contains names of
            other. It is recommended that all input file be kept in the same
            input directory. Each line of projects_xlsx becomes a project_series.
            This is implemented by subclasses.

        enable_cost_and_scaling_modifications : bool
            If True, this method modifies each row of the project list AFTER it has been
            modified by the parameters for to scale certain input values based
            on what has been parametrically modified. This is implemented by subclasses.

        Returns
        -------
        OrderedDict or dict, list
            First element of tuple is an ordered dict that is the result of
            all the runs. Each key is the name of a project and each value
            is the output dictionary of that project. The second element
            is the list of rows for the csv. The third element is the list
            of costs for the spreadsheets. For the first element, serial
            processes implement an OrderedDict (because order can be guaranteed)
            and for the second element processes implement a dict (because
            order cannot be guaranteed).

        Raises
        ------
        NotImplementedError
            NotImplementedError is raised if the method is called on the
            superclass.
        """
        raise NotImplementedError('run_from_project_list_xlsx() can only be called on subclasses')

    def extract_module_type_operation_lists(self, runs_dict):
        """
        This method extract all the cost_by_module_type_operation lists for
        output in an Excel file.

        It finds values for the keys ending in '_module_type_operation'. It
        then concatenates them
        together so they can be easily written to a .csv or .xlsx

        Parameters
        ----------
        runs_dict : dict
            Values are the names of the projects. Keys are the lists of
            dictionaries that are lines for the .csv

        Returns
        -------
        list
            List of dicts to write to the .csv.
        """
        result = []
        for project_results in runs_dict.values():
            for key, value in project_results.items():
                if key.endswith('_module_type_operation'):
                    result.extend(value)
        return result

    def extract_details_lists(self, runs_dict):
        """
        This method extract all .csv lists from the OrderDict of runs to output
        into an Excel or .csv file.

        It finds values for the keys ending in '_csv'. It then concatenates them
        together so they can be easily written to a .csv, .xlsx or other
        columnar format. (The actual writing is left to other functions.

        Parameters
        ----------
        runs_dict : dict
            Values are the names of the projects. Keys are the lists of
            dictionaries that are lines for the .csv

        Returns
        -------
        list
            List of dicts to write to the .csv.
        """
        runs_for_csv = []
        for project_results in runs_dict.values():
            for key, value in project_results.items():
                if key.endswith('_csv'):
                    runs_for_csv.extend(value)
        return runs_for_csv

    def read_project_and_parametric_list_from_xlsx(self):
        """
        This method reads both the project and parametric list from the
        project_list xlsx. It returns them as a tuple.

        To ensure backward compatibility with project list spreadsheets
        that contain only one sheet, the following actions take place

        1. Project list spreadsheet contains one sheet: The only sheet
        that is present is read as the project list, and a dataframe for
        parametrics is returned that has all the columns but no rows so
        it can be used in joins. It doe snot matter what the sheet name
        is.

        2. Project list spreadsheet contains two sheets: The sheet named
        "Project list" is assumed to be the project list and the sheet
        named "Parametric list" is assumed to be the specifications for
        the parametric variables. If one or both of those tab names are
        missing, an exception is raised

        Returns
        -------
        pandas.DataFrame
            The enhanced project list that has support for all parametric
            adjustments for each step.

        Raises
        ------
        KeyError
            When the spreadsheet contains multiple sheets and one or
            both of "Project list" or "Parametric list" are undefined.
        """
        path_to_project_list = self.file_ops.landbosse_input_dir()
        sheets = XlsxDataframeCache.read_all_sheets_from_xlsx('project_list', path_to_project_list)

        # If there is one sheet, make an empty dataframe as a placeholder.
        if len(sheets.values()) == 1:
            first_sheet = list(sheets.values())[0]
            project_list = first_sheet
            parametric_list = pd.DataFrame()

        # If the parametric and project lists exist, read them
        elif 'Parametric list' in sheets.keys() and 'Project list' in sheets.keys():
            project_list = sheets['Project list']
            parametric_list = sheets['Parametric list']

        # Otherwise, raise an exception
        else:
            raise KeyError("Project list needs to have a single sheet or sheets named 'Project list' and 'Parametric list'.")

        # Instantiate and XlsxReader to assemble master input dictionary
        xlsx_reader = XlsxReader()

        # Join in the parametric variable modifications
        parametric_value_list = xlsx_reader.create_parametric_value_list(parametric_list)
        extended_project_list = xlsx_reader.outer_join_projects_to_parametric_values(project_list,
                                                                                 parametric_value_list)

        return extended_project_list



import xlsxwriter
from xlsxwriter.utility import xl_rowcol_to_cell
import pandas as pd
import os
import traceback

from .XlsxFileOperations import XlsxFileOperations


class XlsxGenerator:
    """
    This class is for writing data to Excel files. It is a context manager
    so it used in the following manner:

    with XlsxGenerator(output_xlsx) as xlsx:
        xlsx.joined_with_validation(...)

    Each method call on the context manager makes one or more tabs on the
    output Excel workbook.
    """

    def __init__(self, output_xlsx, file_ops):
        """
        This constructor sets the name of the .xlsx file for writing

        It sets the parameter self.workbook, which is the attribute that
        references the workbook to which all tabs should be written.

        Parameters
        ----------
        output_xlsx : str
            The name of the .xlsx file to write. Do not include the .xlsx at
            the end of the filename. Also, this filename will be timestamped
            before it is written.

        file_ops : XlsxFileOperations
            An instance of XlsxFileOperations to manage file names.
        """

        # Set all instance attributes to None first in the constructor as good
        # coding practice.
        self.workbook = None
        self.header_format = None
        self.scientific_format = None
        self.percent_format = None
        self.output_xlsx_path = os.path.join(file_ops.landbosse_output_dir(), f'{output_xlsx}.xlsx')
        self.file_ops = file_ops

    @classmethod
    def write_project_data(cls, project_data_dataframes, project_data_output_xlsx_path):
        """
        This writes a dictionary full of dataframes to an Excel spreadsheet.

        There is no special formatting done to the output Excel sheets.

        This is a class method so an instance of this class does not need
        to be created to write project data. This enables this method to
        be nicely accessed from XlsxManagerRunner subclasses.

        Parameters
        ----------
        project_data_dataframes : dict
            The dataframes to write to the output sheets. The key names will
            be used as the sheet names.

        project_data_output_xlsx_path : str
            The absolute pathname to write the .xlsx file to.
        """
        with pd.ExcelWriter(project_data_output_xlsx_path, mode='w') as writer:
            for name, df in project_data_dataframes.items():
                df.to_excel(writer, sheet_name=name)

    def __enter__(self):
        """
        Opens the workbook for writing and sets the formatting.

        Returns
        -------
        self
            Returns self for easy use in the context manager.
        """
        self.workbook = xlsxwriter.Workbook(self.output_xlsx_path, {'nan_inf_to_errors': True})
        self.set_workbook_formats()
        return self

    def __exit__(self, exception_type, exception_val, exception_traceback):
        """
        Closes the workbook

        Parameters
        ----------
        exception_type
            The type of the exception, if there is one.
        exception_val
            The value of the exception, if there is one.
        exception_traceback
            Traceback of the exception problem.

        Returns
        -------
        bool
            True if the workbook is closed properly. False if something
            prevented normal closure.
        """
        if exception_type:
            print('exception_type: {}'.format(exception_type))
            print('exception_val: {}'.format(exception_val))
            print('exception_traceback:')
            traceback.print_tb(exception_traceback)
        self.workbook.close()
        return True

    def set_workbook_formats(self):
        """
        This method creates the formats in the workbook. It is called upon entry
        into the context manager.
        """
        self.header_format = self.workbook.add_format()
        self.header_format.set_bold()
        self.header_format.set_text_wrap()
        self.scientific_format = self.workbook.add_format()
        self.scientific_format.set_num_format('0.00E+00')
        self.scientific_format.set_align('left')
        self.percent_format = self.workbook.add_format()
        self.percent_format.set_num_format('0.0%')
        self.percent_format.set_align('left')
        self.accounting_format = self.workbook.add_format()
        self.accounting_format.set_num_format('$ #,##0')

    def tab_costs_by_module_type_operation(self, rows):
        """
        This writes the costs_by_module_type_operation tab.

        Parameters
        ----------
        rows : list
            List of dictionaries that are each row in the output
            sheet.
        """
        worksheet = self.workbook.add_worksheet('costs_by_module_type_operation')
        for idx, col_name in enumerate(['Project ID with serial',
                                        'Number of turbines',
                                        'Turbine rating MW',
                                        'Rotor diameter m',
                                        'Module',
                                        'Operation ID',
                                        'Type of cost',
                                        'Cost per turbine',
                                        'Cost per project',
                                        'USD/kW per project']):
            worksheet.write(0, idx, col_name, self.header_format)
        for row_idx, row in enumerate(rows):
            worksheet.write(row_idx + 1, 0, row['project_id_with_serial'])
            worksheet.write(row_idx + 1, 1, row['num_turbines'])
            worksheet.write(row_idx + 1, 2, row['turbine_rating_MW'])
            worksheet.write(row_idx + 1, 3, row['rotor_diameter_m'])
            worksheet.write(row_idx + 1, 4, row['module'])
            worksheet.write(row_idx + 1, 5, row['operation_id'])
            worksheet.write(row_idx + 1, 6, row['type_of_cost'])
            worksheet.write(row_idx + 1, 7, row['cost_per_turbine'], self.accounting_format)
            worksheet.write(row_idx + 1, 8, row['cost_per_project'], self.accounting_format)
            worksheet.write(row_idx + 1, 9, row['usd_per_kw_per_project'], self.accounting_format)
            worksheet.set_column(0, 5, 25)
            worksheet.set_column(6, 10, 17)
        worksheet.freeze_panes(1, 0)  # Freeze the first row.

    def tab_details(self, rows):
        """
        This writes a detailed outputs tab. It takes a list of dictionaries
        as the parameters and in each of those dictionaries it looks at the keys:

        ['project_id', 'module', 'type', 'variable_df_key_col_name', 'unit', 'numeric value', 'non_numeric_value']

        The values of each of those keys become each cell in the row.

        Parameters
        ----------
        rows : list
            list of dicts. See above.
        """
        worksheet = self.workbook.add_worksheet('details')
        worksheet.set_column(3, 3, 66)
        worksheet.set_column(4, 4, 17)
        worksheet.set_column(5, 5, 66)
        worksheet.set_column(0, 2, 17)

        for idx, col_name in enumerate(['Project ID with serial', 'Module', 'Variable or DataFrame', 'name', 'unit', 'Numeric value', 'Non-numeric value']):
            worksheet.write(0, idx, col_name, self.header_format)

        # Go through each row and create Excel rows from each of those rows.
        for row_idx, row in enumerate(rows):
            worksheet.write(row_idx + 1, 0, row['project_id_with_serial'])
            worksheet.write(row_idx + 1, 1, row['module'])
            worksheet.write(row_idx + 1, 2, row['type'])
            worksheet.write(row_idx + 1, 3, row['variable_df_key_col_name'])
            worksheet.write(row_idx + 1, 4, row['unit'])

            value = row['value']
            value_is_number = self._is_numeric(value)
            if value_is_number:
                worksheet.write(row_idx + 1, 5, value, self.scientific_format)
            else:
                worksheet.write(row_idx + 1, 6, value)

            # If there is a last_number, which means this is a dataframe row that has a number
            # at the end, write this into the numeric value column. This overrides automatic
            # type detection.

            if 'last_number' in row:
                worksheet.write(row_idx + 1, 5, row['last_number'], self.scientific_format)

            # Certain data are pairs of numeric and non-numeric values. If a key of
            # "non_numeric_value" exists, put that in column 6.
            # An example is mobilization of an LB75-SL3F-Offload at some numeric cost

            if 'non_numeric_value' in row:
                worksheet.write(row_idx + 1, 6, row['non_numeric_value'])

        worksheet.freeze_panes(1, 0)  # Freeze the first row.

    def _is_numeric(self, value):
        """
        This method tests if a value is a numeric (that is, can be parsed
        by float()) or non numeric (which cannot be parsed).

        The decision from this method determines whether values go into
        the numeric or non-numeric columns.

        Parameters
        ----------
        value
            The value to be tested.

        Returns
        -------
        bool
            True if the value is numeric, False otherwise.
        """
        try:
            float(value)
        except ValueError:
            return False
        return True



import re

import pandas as pd
import numpy as np
from math import ceil

from .XlsxOperationException import XlsxOperationException
from .WeatherWindowCSVReader import read_weather_window, extend_weather_window
from ..model import DefaultMasterInputDict
from .GridSearchTree import GridSearchTree


class XlsxReader:
    """
    This class is for reading input data from .xlsx files.

    There are two sets of data to be read from .xlsx files. The first set
    of data is read from the following sheets:

    - components

    - cable_specs

    - equip

    - crane_specs

    - development

    - crew_price

    - crew

    - equip_price

    - material_price

    - rsmeans

    - site_facility_building_area

    - weather_window

    The second set of data are read from a single sheet as described below.

    The first set of data represent a database used by the various modules. Queries
    on these data allow calculation based on labor, material, crane capacity
    and so on. These data can be shared among different projects. These data
    are called project_data.

    The second set of data represent the parameters specific to each project.
    These parameters include things values like hub height, rotor diameter,
    etc. These data are particular to a single project. These data are
    referred to as the project.

    This class also handles modifications of project data dataframes for
    parametric runs and all the other dataframe manipulations that make that
    possible.
    """

    def create_parametric_value_list(self, parametric_list):
        """
        Assuming we have a "Parametric list" sheet/dataframe like the following

        | Project ID | Dataframe name       | Row name | Column name | Start | End | Step
        |------------|----------------------|----------|-------------|-------|-----|------
        | project1   | alpha                | fizz     | buzz        | 0     | 12  | 6
        | project1   | beta                 | foo      | bar         | 0     | 12  | 6
        | project2   | gamma                | dogs     | cats        | 21    | 27  | 3

        Translate this data frame into a data frame of the following format:

        | Project ID | Serial      | alpha/fizz/buzz | beta/foo/bar | gamma/dogs/cats
        |------------|-------------|-----------------|--------------|----------------|
        | project1   | project1_00 | 0               | 0            | NaN            |
        | project1   | project1_01 | 0               | 6            | NaN            |
        | project1   | project1_02 | 0               | 12           | NaN            |
        | project1   | project1_03 | 6               | 0            | NaN            |
        | project1   | project1_04 | 6               | 6            | NaN            |
        | project1   | project1_05 | 6               | 12           | NaN            |
        | project1   | project1_06 | 12              | 0            | NaN            |
        | project1   | project1_07 | 12              | 6            | NaN            |
        | project1   | project1_08 | 12              | 12           | NaN            |
        | project2   | project2_09 | NaN             | NaN          | 21             |
        | project2   | project2_10 | NaN             | NaN          | 24             |
        | project2   | project2_11 | NaN             | NaN          | 27             |

        Here NaN is not an error. It is the normal way Pandas handles
        empty cells It means that, for a particular project,
        to the dataframe/row/column is needed and that the value in that
        dataframe cell should remain unchanged.

        Also, note that the serial numbers are strings that should be left
        padded with zeros. The left padding means that when the strings are
        sorted alphabetically, they will end up in the same order as numeric
        sorting.

        Parameters
        ----------
        parametric_list : pandas.DataFrame
            The first dataframe shown above.

        Returns
        -------
        pandas.DataFrame
            The second dataframe shown above.
        """

        # If the parametric_list is empty, this probably means the sheet
        # that had the parametric values is empty.
        if parametric_list.empty:
            return pd.DataFrame([
                {
                    'Project ID': None,
                    'Project ID with serial': None
                }
            ])

        all_parametric_value_rows = []

        # Group all the projects by their ID and iterate over each group
        group_by_project = parametric_list.groupby('Project ID')
        for name, group in group_by_project:

            # For each group/project ID, build a new grid search tree
            grid_search_tree = GridSearchTree(group)
            grid = grid_search_tree.build_grid_tree_and_return_grid()

            # After the tree is built, make new rows that have individual
            # parametric values.
            #
            # A grid_point is set of cell specifications and their values.
            # given our example above, a grid point would be a list of
            # dictionaries about dataframe cells to modify.
            #
            # Grid point 0: alpha/fizz/buzz = 0, beta/foo/bar = 0
            # Grid point 1: alpha/fizz/buzz = 0, beta/foo/bar = 6
            # etc
            parametric_value_rows = []
            for grid_point in grid:
                parametric_value_row = dict()
                for axis in grid_point:
                    parametric_value_row[axis['cell_specification']] = axis['value']
                parametric_value_rows.append(parametric_value_row)

            # Assign the project ID to each row. Because of the grouping,
            # tha name is the same as the project ID.
            for parametric_value_row in parametric_value_rows:
                parametric_value_row['Project ID'] = name

            # Add all these new rows into the master list before
            # iterating again.
            all_parametric_value_rows.extend(parametric_value_rows)

        # Create project names with serial numbers for each row
        for index, parametric_value_row in enumerate(all_parametric_value_rows):
            project_id = parametric_value_row['Project ID']
            serial = self.create_serial_number(project_id, index, len(all_parametric_value_rows))
            parametric_value_row['Project ID with serial'] = serial

        # Make a dataframe out of the list of dictionaries. This will
        # add NaN where appropriate

        result = pd.DataFrame(all_parametric_value_rows)

        return result

    def outer_join_projects_to_parametric_values(self, project_list, parametric_value_list):
        """
        Consider the dataframe we made in create_parametric_value_list.
        Call it parametric_value_list:

        | Project ID | Project ID with serial      | alpha/fizz/buzz | beta/foo/bar | gamma/dogs/cats
        |------------|-----------------------------|-----------------|--------------|----------------|
        | project1   | project1_00                 | 0               | 0            | NaN            |
        | project1   | project1_01                 | 0               | 6            | NaN            |
        | project1   | project1_02                 | 0               | 12           | NaN            |
        | project1   | project1_03                 | 6               | 0            | NaN            |
        | project1   | project1_04                 | 6               | 6            | NaN            |
        | project1   | project1_05                 | 6               | 12           | NaN            |
        | project1   | project1_06                 | 12              | 0            | NaN            |
        | project1   | project1_07                 | 12              | 6            | NaN            |
        | project1   | project1_08                 | 12              | 12           | NaN            |
        | project2   | project2_09                 | NaN             | NaN          | 21             |
        | project2   | project2_10                 | NaN             | NaN          | 24             |
        | project2   | project2_11                 | NaN             | NaN          | 27             |

        Now consider that we have the following project list data frame called
        project_list:

        Project ID | Project data file | Total project construction time months | ...
        -----------|-------------------|----------------------------------------|---
        project1   | project1_data     | 9                                      | ...
        project2   | project2_data     | 9                                      | ...
        project3   | project3_data     | 9                                      | ...

        OUTER join project_list with parametric_value_list such that you
        get the following data frame. The outer join ensures that even
        projects that do not have parametrics attached to them find
        their way to the list of projects to be executed.

        | Project ID | Project ID with serial      | alpha/fizz/buzz | beta/foo/bar | gamma/dogs/cats  Project data file | ...
        |------------|-----------------------------|-----------------|--------------|----------------|-------------------| ...
        | project1   | project1_00                 | 0               | 0            | NaN            | project1_data     | ...
        | project1   | project1_01                 | 0               | 6            | NaN            | project1_data     | ...
        | project1   | project1_02                 | 0               | 12           | NaN            | project1_data     | ...
        | project1   | project1_03                 | 6               | 0            | NaN            | project1_data     | ...
        | project1   | project1_04                 | 6               | 6            | NaN            | project1_data     | ...
        | project1   | project1_05                 | 6               | 12           | NaN            | project1_data     | ...
        | project1   | project1_06                 | 12              | 0            | NaN            | project1_data     | ...
        | project1   | project1_07                 | 12              | 6            | NaN            | project1_data     | ...
        | project1   | project1_08                 | 12              | 12           | NaN            | project1_data     | ...
        | project2   | project2_09                 | NaN             | NaN          | 21             | project2_data     | ...
        | project2   | project2_10                 | NaN             | NaN          | 24             | project2_data     | ...
        | project2   | project2_11                 | NaN             | NaN          | 27             | project2_data     | ...
        | project3   | NaN                         | NaN             | NaN          | NaN            | project3_data     | ...

        This data frame is in place to run with a modified project manager
        runner. The modification to the runner will run modifications to the
        project data dataframes before running them in a project. The columns
        that contain the modifications are the ones that contain the slashes.

        The outer join will ensure that all projects, including those that do
        not have parametric modification, are joined in. Further, the outer
        join will mean that there are no additional flags that can introduce
        errors in the project list.

        As before, NaN values are expected and signify that a particular
        parametric variable does not need to be modified for the project
        that is in the same row.

        Also, NaN in the Serial column means that a serial number will be
        added later because the project did not have any parametric modifications
        """
        result = project_list.merge(right=parametric_value_list, how='left', on='Project ID')
        return result

    def modify_project_data_and_project_list(self, project_data_dataframes, project_parameters):
        """
        This method modifies project data dataframes according to the
        parametric modifications in the project parameters. It does not
        return a value because dataframes are modified in place.

        Note: This method will modify the dataframes in place. This isn't
        a problem if the dataframes are returned from
        XlsxReader.read_all_sheets_from_xlsx.

        If the dataframe name, column name or row name are not found, an
        XlsxOperationException is raised.

        Also, it modifies (once again in plance) the project parameters
        according to the parametrics.

        Parameters
        ----------
        project_data_dataframes : dict
            Keys in this dictionary are the names of the sheets where
            the dataframes are parsed from. Values are the dataframes
            to be modified.

        project_parameters : pandas.Series
            The enhanced project parameters as created by
            create_parametric_value_list that have the values to
            placed into the dataframes.

        Returns
        -------
        None
            Dataframes are modified in place.

        Raises
        ------
        XlsxOperationException
            This exception is raised of a dataframe, row or column
            is not found. The message is descriptive to help diagnose the
            problem during operation.
        """
        # This is a regex to match a column name that specifies a change to make
        # to a cell
        cell_spec_re = re.compile('^.*/.*/.*$')

        # Go through each project parameter
        for index, value in project_parameters.iteritems():

            # If the column specifies a cell to change in the dataframe
            # or project list, inspect it to ensure it points somewhere
            # valid and change the project list or the dataframe

            if cell_spec_re.match(index) and not pd.isnull(value):
                dataframe_name, row_name, column_name = index.split('/')

                # First, branch on whether this a project list parameter
                # or a dataframe parameter. First, if it is a project list
                # parameter.
                if dataframe_name == 'project list':
                    if column_name not in project_parameters:
                        raise XlsxOperationException(
                            f'Column {column_name} not found in project parameters'
                        )
                    if not pd.isnull(value):
                        project_parameters[column_name] = value

                # Second, if it is a dataframe parameter
                else:
                    # Check if dataframe exists
                    if dataframe_name not in project_data_dataframes:
                        raise XlsxOperationException(
                            f'Datframe {dataframe_name} not found. Please check the project_data spreadsheet and project_list.')

                    df = project_data_dataframes[dataframe_name]
                    first_col = df.columns[0]

                    # Check if row exists
                    if df.loc[df[first_col] == row_name].empty:
                        raise XlsxOperationException(
                            f'Row {row_name} not found in dataframe {dataframe_name}. Please check the project_data spreadsheet and project_list.')

                    # Check if column exists
                    if df.loc[df[first_col] == row_name, column_name].empty:
                        raise XlsxOperationException(
                            f'Column {column_name} not found in dataframe {dataframe_name}. Please check the project_data spreadsheet and project_list.')

                    # If all the above check pass, check to make sure the value is not nan.
                    # If it is not nan, then a modification needs to be made, either to the
                    # dataframe or the project list.
                    if not pd.isnull(value):
                        df.loc[df[first_col] == row_name, column_name] = value

    def create_master_input_dictionary(self, project_data_dataframes, project_parameters):
        """
        This method takes a dictionary of dataframes that are the project data
        and unites them with the project parameters as found in the project list
        sheet.

        Parameters
        ----------
        project_data_dataframes : dict
            This is a dictionary for the project data .xlsx file. The keys
            are names of sheets and the values are the dataframe contents of
            the sheets.

        project_parameters : pandas.Series
            Series representing the project data, which is the second set
            of data described in the XlsxReader class docstring. The caller
            of this function is responsible for parsing out the project
            data into a series from which these data can be extracted.
            See the subclasses of XlsxManagerRunner for examples on how this
            project series is read from a spreadsheet.

        Returns
        -------
        dict
            An master input dictionary suitable to pass to an instance
            of Manager to run all the cost module sin LandBOSSE.
        """
        # First, read all inputs that come from .csv or .xlsx files.
        # erection_input_worksheets come from the input data spreadsheet.
        # Their string values are the names of the sheets in the Excel
        # workbook and the keys in the erection_project_data_dict dictionary.

        # Incomplete project dict will hold the input dictionary
        # configurations.
        incomplete_input_dict = dict()
        incomplete_input_dict['error'] = dict()

        # Read all project_data sheets.
        # The erection module takes in a bunch of keys and values under the
        # 'project_data' key in the incomplete_input_dict

        # Apply the labor multipliers
        incomplete_input_dict['labor_cost_multiplier'] = project_parameters['Labor cost multiplier']
        labor_cost_multiplier = incomplete_input_dict['labor_cost_multiplier']
        self.apply_labor_multiplier_to_project_data_dict(project_data_dataframes, labor_cost_multiplier)

        erection_input_worksheets = [
            'crane_specs',
            'equip',
            'crew',
            'equip_price',
            'crew_price',
            'material_price',
            'components'
        ]

        erection_project_data_dict = dict()
        for worksheet in erection_input_worksheets:
            erection_project_data_dict[worksheet] = project_data_dataframes[worksheet]

        # Add the erection project data to the incomplete_input_dict
        incomplete_input_dict['project_data'] = erection_project_data_dict

        # Get the first set of data
        incomplete_input_dict['rsmeans'] = project_data_dataframes['rsmeans']
        incomplete_input_dict['site_facility_building_area_df'] = project_data_dataframes['site_facility_building_area']
        incomplete_input_dict['material_price'] = project_data_dataframes['material_price']

        # Read development tab, if it exists (it is optional since development costs can
        # be placed in the project list):
        if 'development' in project_data_dataframes:
            incomplete_input_dict['development_df'] = project_data_dataframes['development']

        # FoundationCost needs to have all the component data split into separate
        # NumPy arrays.
        incomplete_input_dict['component_data'] = erection_project_data_dict['components']
        for component in incomplete_input_dict['component_data'].keys():
            incomplete_input_dict[component] = np.array(incomplete_input_dict['component_data'][component])

        incomplete_input_dict['cable_specs_pd'] = project_data_dataframes['cable_specs']

        # For development cost, legacy input data will specify an itemized
        # breakdown in the project data. Newer input data will specify the
        # labor cost in the project list.
        #
        # In the DevelopmentCost module, this change will be detected by the
        # absence of a development_labor_cost_usd key in the master input
        # dictionary. In that case, the development cost will be pulled from
        # the prject data.
        if 'Development labor cost USD' in project_parameters:
            incomplete_input_dict['development_labor_cost_usd'] = project_parameters['Development labor cost USD']

        # These columns come from the columns in the project definition .xlsx
        incomplete_input_dict['project_id'] = project_parameters['Project ID']
        incomplete_input_dict['crane_breakdown_fraction'] = project_parameters['Crane breakdown fraction']
        incomplete_input_dict['num_turbines'] = project_parameters['Number of turbines']
        incomplete_input_dict['construct_duration'] = project_parameters['Total project construction time (months)']
        incomplete_input_dict['hub_height_meters'] = project_parameters['Hub height m']
        incomplete_input_dict['rotor_diameter_m'] = project_parameters['Rotor diameter m']
        incomplete_input_dict['wind_shear_exponent'] = project_parameters['Wind shear exponent']
        incomplete_input_dict['turbine_rating_MW'] = project_parameters['Turbine rating MW']
        incomplete_input_dict['breakpoint_between_base_and_topping_percent'] = \
            project_parameters['Breakpoint between base and topping (percent)']
        incomplete_input_dict['fuel_usd_per_gal'] = project_parameters['Fuel cost USD per gal']
        incomplete_input_dict['rate_of_deliveries'] = project_parameters['Rate of deliveries (turbines per week)']
        incomplete_input_dict['turbine_spacing_rotor_diameters'] = project_parameters['Turbine spacing (times rotor diameter)']
        incomplete_input_dict['depth'] = project_parameters['Foundation depth m']
        incomplete_input_dict['rated_thrust_N'] = project_parameters['Rated Thrust (N)']
        incomplete_input_dict['bearing_pressure_n_m2'] = project_parameters['Bearing Pressure (n/m2)']
        incomplete_input_dict['gust_velocity_m_per_s'] = project_parameters['50-year Gust Velocity (m/s)']
        incomplete_input_dict['project_size_megawatts'] = project_parameters['Number of turbines'] * project_parameters['Turbine rating MW']

        if project_parameters['Calculate road cost for distributed wind? (y/n)'] == 'y':
            incomplete_input_dict['road_distributed_wind'] = True
        else:
            incomplete_input_dict['road_distributed_wind'] = False
        incomplete_input_dict['site_prep_area_m2'] = project_parameters['Site prep area for Distributed wind (m2)']
        incomplete_input_dict['road_length_adder_m'] = project_parameters['Road length adder (m)']
        incomplete_input_dict['fraction_new_roads'] = project_parameters['Percent of roads that will be constructed']
        incomplete_input_dict['road_quality'] = project_parameters['Road Quality (0-1)']
        incomplete_input_dict['line_frequency_hz'] = project_parameters['Line Frequency (Hz)']
        incomplete_input_dict['plant_capacity_MW'] = project_parameters['Turbine rating MW'] * project_parameters['Number of turbines']
        incomplete_input_dict['row_spacing_rotor_diameters'] = project_parameters['Row spacing (times rotor diameter)']
        incomplete_input_dict['user_defined_distance_to_grid_connection'] = project_parameters['Flag for user-defined home run trench length (0 = no; 1 = yes)']
        incomplete_input_dict['distance_to_grid_connection_km'] = project_parameters['Combined Homerun Trench Length to Substation (km)']
        incomplete_input_dict['crew'] = incomplete_input_dict['project_data']['crew']
        incomplete_input_dict['crew_cost'] = incomplete_input_dict['project_data']['crew_price']

        #read in RSMeans per diem:
        crew_cost = incomplete_input_dict['project_data']['crew_price']
        crew_cost = crew_cost.set_index("Labor type ID", drop=False)
        incomplete_input_dict['rsmeans_per_diem'] = crew_cost.loc['RSMeans', 'Per diem USD per day']

        incomplete_input_dict['fuel_cost_usd_per_gal'] = project_parameters['Fuel cost USD per gal']

        incomplete_input_dict['line_frequency_hz'] = project_parameters['Line Frequency (Hz)']
        incomplete_input_dict['plant_capacity_MW'] = project_parameters['Turbine rating MW'] * project_parameters['Number of turbines']
        incomplete_input_dict['row_spacing_rotor_diameters'] = project_parameters['Row spacing (times rotor diameter)']
        incomplete_input_dict['user_defined_home_run_trench'] = project_parameters[
            'Flag for user-defined home run trench length (0 = no; 1 = yes)']
        incomplete_input_dict['trench_len_to_substation_km'] = project_parameters[
            'Combined Homerun Trench Length to Substation (km)']

        # Add inputs for transmission & Substation modules:
        incomplete_input_dict['distance_to_interconnect_mi'] = project_parameters['Distance to interconnect (miles)']
        incomplete_input_dict['interconnect_voltage_kV'] = project_parameters['Interconnect Voltage (kV)']
        new_switchyard = True
        if project_parameters['New Switchyard (y/n)'] == 'y':
            new_switchyard = True
        else:
            new_switchyard = False
        incomplete_input_dict['new_switchyard'] = new_switchyard

        incomplete_input_dict['critical_speed_non_erection_wind_delays_m_per_s'] = project_parameters['Non-Erection Wind Delay Critical Speed (m/s)']
        incomplete_input_dict['critical_height_non_erection_wind_delays_m'] = project_parameters['Non-Erection Wind Delay Critical Height (m)']

        incomplete_input_dict['road_width_ft'] = project_parameters['Road width (ft)']
        incomplete_input_dict['road_thickness'] = project_parameters['Road thickness (in)']
        incomplete_input_dict['crane_width'] = project_parameters['Crane width (m)']
        incomplete_input_dict['num_hwy_permits'] = project_parameters['Number of highway permits']
        incomplete_input_dict['num_access_roads'] = project_parameters['Number of access roads']
        incomplete_input_dict['overtime_multiplier'] = project_parameters['Overtime multiplier']
        incomplete_input_dict['allow_same_flag'] = True if project_parameters['Allow same flag'] == 'y' else False

        override_total_mgmt_cost_col_name = 'Override total management cost for distributed (0 does not override)'
        if override_total_mgmt_cost_col_name in project_parameters and project_parameters[override_total_mgmt_cost_col_name] > 0:
            incomplete_input_dict['override_total_management_cost'] = \
                project_parameters[override_total_mgmt_cost_col_name]
        else:
            incomplete_input_dict['markup_contingency'] = project_parameters['Markup contingency']
            incomplete_input_dict['markup_warranty_management'] = project_parameters['Markup warranty management']
            incomplete_input_dict['markup_sales_and_use_tax'] = project_parameters['Markup sales and use tax']
            incomplete_input_dict['markup_overhead'] = project_parameters['Markup overhead']
            incomplete_input_dict['markup_profit_margin'] = project_parameters['Markup profit margin']

        # The weather window is stored on a sheet of the project_data, but
        # needs preprocessing after it is read. The preprocessing changes it
        # from wind toolkit format to a dataframe.
        number_of_months_for_construction = int(project_parameters['Total project construction time (months)'])
        weather_window_input = project_data_dataframes['weather_window']
        weather_window_intermediate = read_weather_window(weather_window_input)
        extended_weather_window = extend_weather_window(weather_window_intermediate, number_of_months_for_construction)
        incomplete_input_dict['weather_window'] = extended_weather_window

        # Now fill any missing values with sensible defaults.
        defaults = DefaultMasterInputDict()
        master_input_dict = defaults.populate_input_dict(incomplete_input_dict=incomplete_input_dict)
        return master_input_dict

    def apply_labor_multiplier_to_project_data_dict(self, project_data_dict, labor_cost_multiplier):
        """
        Applies the labor multiplier to the dataframes that contain the labor
        costs in the project_data_dict. The dataframes are values in the
        dictionary. The keys are "crew_price" and "rsmeans".

        For the crew_price dataframe, the labor_cost_multiplier is broadcast
        down the "Hourly rate USD per hour" and the "Per diem USD per day" columns.

        For the rsmeans dataframe, rows that have "Labor" for the "Type of cost"
        column are found and, for those rows, the values in the "Rate USD per unit"
        column is multiplied by the multiplier

        The dataframes are modified in place.

        Parameters
        ----------
        project_data_dict : dict
            The dictionary that has the dataframes as values.

        labor_cost_multiplier : float
            The scalar labor cost multiplier.
        """
        crew_price = project_data_dict['crew_price']
        crew_price_new_hourly_rates = crew_price['Hourly rate USD per hour'] * labor_cost_multiplier
        crew_price_new_per_diem_rates = crew_price['Per diem USD per day'] * labor_cost_multiplier
        crew_price['Hourly rate USD per hour'] = crew_price_new_hourly_rates
        crew_price['Per diem USD per day'] = crew_price_new_per_diem_rates

        rsmeans = project_data_dict['rsmeans']

        # This function maps new labor rates in rsmeans. Used with an apply()
        # call, it will create a new Series that can be mapped back onto the
        # original dataframe.
        #
        # If the column "Type of cost" is "Labor" then return the current cost
        # times the labor multiplier. If "Type of cost" isn't "Labor" then
        # just return the current cost.
        def map_labor_rates(row):
            if row['Type of cost'] == 'Labor':
                return row['Rate USD per unit'] * labor_cost_multiplier
            else:
                return row['Rate USD per unit']

        rsmeans_new_labor_rates = rsmeans.apply(map_labor_rates, axis=1)
        rsmeans.drop(columns=['Rate USD per unit'], inplace=True)
        rsmeans['Rate USD per unit'] = rsmeans_new_labor_rates

    def apply_cost_and_scaling_modifications_to_project_parameters(self, project_parameters):
        """
        This applies the cost and scaling modification to project parameters
        specified in the project list.

        Note: It is meant to be called on the parameter list AFTER it has been
        modified with the parametrics by modify_project_data_and_project_list()
        above.

        It modifies the parameters IN PLACE.

        Parameters
        ----------
        project_parameters : pd.Series
            The project parameters to be modified.
        """
        project_size_MW = project_parameters['Number of turbines'] * project_parameters['Turbine rating MW']
        hub_height_m = project_parameters['Hub height m']
        flag_use_user_homerun = project_parameters['Flag for user-defined home run trench length (0 = no; 1 = yes)']
        nameplate = project_parameters['Turbine rating MW']

        distance_to_interconnect_mi = 0.0 if project_size_MW <= 20 else (0.009375 * project_size_MW + 0.625)
        interconnect_voltage_kV = 0.4398 * project_size_MW + 60.204
        new_switchyard_y_n = 'n' if project_size_MW <= 40 else 'y'
        road_length_adder_m = 1e3 if project_size_MW <= 20 else (13.542 * project_size_MW + 1458.3)

        # if greater than 20 MW, then breakpoint between base and topping at 35 meters
        breakpoint_between_base_and_topping = 0.0 if project_size_MW <= 20 else (35 / hub_height_m)

        number_of_access_roads = 0.0 if project_size_MW <= 20 else ceil(0.0052 * project_size_MW + 0.7917)
        number_of_highway_permits = ceil(0.2 * project_parameters['Number of turbines'])
        if flag_use_user_homerun is 1:
             project_parameters['Combined Homerun Trench Length to Substation (km)'] = 0.1776 * project_size_MW - 2.551

        # 10 deliveries per week for 1.5 MW machines
        rate_deliveries = ceil(15 / nameplate)

        # $17,000 / MW for a development cost estimate
        development_labor_cost_usd = project_size_MW * 17000

        project_parameters['Rate of deliveries(turbines per week)'] = rate_deliveries
        project_parameters['Development labor cost USD'] = development_labor_cost_usd
        project_parameters['Project size MW'] = project_size_MW
        project_parameters['Distance to interconnect (miles)'] = distance_to_interconnect_mi
        project_parameters['Interconnect Voltage (kV)'] = interconnect_voltage_kV
        project_parameters['New Switchyard (y/n)'] = new_switchyard_y_n
        project_parameters['Road length adder (m)'] = road_length_adder_m
        project_parameters['Breakpoint between base and topping (percent)'] = breakpoint_between_base_and_topping
        project_parameters['Number of access roads'] = number_of_access_roads
        project_parameters['Number of highway permits'] = number_of_highway_permits

    def create_serial_number(self, project_id, index, max_index):
        """
        create_serial_number creates serial numbers left padded with
        zeros. By left padding the numbers, alphabetic sorts and numeric
        sorts create the same sequence. For example, in a sequence of
        100 total elements, 5 would be represented as "005"

        This left padded number is appended to the end of a string so
        you have something like "project_001"

        Parameters
        ----------
        project_id : str
            Base name of the project.

        index : int
            The index of the project in the sequence of projects

        max_index : int
            The total number of indices in the sequence.

        Returns
        -------
        str
            The left padded serial number as a string.
        """
        total_digit_count = 1
        index_digit_count = len(str(index))

        if max_index < 10:
            total_digit_count = 1
        elif 0 < max_index < 1e1 - 1:
            total_digit_count = 1
        elif 1e1 <= max_index < 1e2 - 1:
            total_digit_count = 2
        elif 1e2 <= max_index < 1e3 - 1:
            total_digit_count = 3
        elif 1e3 <= max_index < 1e4 - 1:
            total_digit_count = 4
        elif 1e4 <= max_index < 1e5 - 1:
            total_digit_count = 5
        elif 1e5 <= max_index < 1e6 - 1:
            total_digit_count = 6
        elif 1e6 <= max_index < 1e7 - 1:
            total_digit_count = 7
        elif 1e7 <= max_index < 1e8 - 1:
            total_digit_count = 8
        else:
            total_digit_count = 9

        padding = '0' * (total_digit_count - index_digit_count)
        return f'{project_id}_{padding}{index}'



import os
import sys
from datetime import datetime
from shutil import copy2
from shutil import copytree

from .XlsxOperationException import XlsxOperationException


class XlsxFileOperations:
    """
    This class is made to handle file naming and copying.
    """

    def __init__(self):
        """
        The __init__() method just makes a timestamp that will be used throughout
        the lifetime of this instance.
        """
        dt = datetime.now()
        self.timestamp = f'{dt.year}-{dt.month}-{dt.day}-{dt.hour}-{dt.minute}-{dt.second}'

    def get_input_output_paths_from_argv_or_env(self):
        """
        This uses the sys.argv object to inspect the command line to find input
        and output paths as specified on the command line. It expects the
        command line to have options in the following form:

        -i [input file] -o [output file]

        If one or both of these is missing, then it is filled with the defaults
        from the environment variables LANDBOSSE_INPUT_DIR or LANDBOSSE_OUTPUT_DIR

        If all of them are missing the method defaults to 'inputs/' and 'outputs/'

        There is a third option this method looks for:

        --validate or -v

        If that parameter is present, then --output or -o is optional. If enabled,
        validation mode will accept an input directory with --input that must also have a
        landbosse-output.xlsx file. The model will then run on the inputs. Instead
        of writing a file of the output, the output will stay in memory. The
        landbosse-output.xlsx will then be loaded and compared against the in
        memory output. If the outputs are the same, the validation passes
        because the results were reproduced. Otherwise, the validation failed
        because something broke in the model.

        There is a fourth option this method looks for:

        --scaling or -s

        This enables auto scaling of input parameters as found in XlsxReader's
        apply_cost_and_scaling_modifications_to_project_parameters() method.

        Parameters
        ----------
        This function takes no parameters.

        Returns
        -------
        str, str, bool
            The first two strings are paths to the input and output files
            respectively. The third bool is True if validation mode is
            enabled, False for normal operation.
        """

        # Get the fallback paths from the environment variables and set their
        # defaults. These defualts are used

        input_path_from_env = os.environ.get('LANDBOSSE_INPUT_DIR', 'input')
        output_path_from_env = os.environ.get('LANDBOSSE_OUTPUT_DIR', 'output')

        # input and output paths from command line are initially set to None
        # to indicate they have not been found yet.

        input_path_from_arg = None
        output_path_from_arg = None

        # This is for validation option detection
        validation_enabled =  '--validate' in sys.argv or '-v' in sys.argv

        # This is for scaling study operation
        enable_scaling_study = '--scaling' in sys.argv or '-s' in sys.argv

        # If validation and scaling study are simultaneously enabled, raise
        # an error

        if validation_enabled and enable_scaling_study:
            raise XlsxOperationException('--scaling-study and --validate cannot be enabled at the same time.')

        # Look for the input path on command line
        if '--input' in sys.argv and sys.argv.index('--input') + 1 < len(sys.argv):
            input_idx = sys.argv.index('--input') + 1
            input_path_from_arg = sys.argv[input_idx]

        if '-i' in sys.argv and sys.argv.index('-i') + 1 < len(sys.argv):
            input_idx = sys.argv.index('-i') + 1
            input_path_from_arg = sys.argv[input_idx]

        # Look for the output path on command line
        if '--output' in sys.argv and sys.argv.index('--output') + 1 < len(sys.argv):
            output_idx = sys.argv.index('--output') + 1
            output_path_from_arg = sys.argv[output_idx]

        if '-o' in sys.argv and sys.argv.index('-o') + 1 < len(sys.argv):
            output_idx = sys.argv.index('-o') + 1
            output_path_from_arg = sys.argv[output_idx]

        # Find the final input and output paths. If a command line argument was
        # found for input and/or output, that is used. If it wasn't found,
        # the value from the environment variable search is returned, which includes
        # the default if the environment variable itself wasn't found.

        input_path = input_path_from_arg if input_path_from_arg is not None else input_path_from_env
        output_path = output_path_from_arg if output_path_from_arg is not None else output_path_from_env

        # Return the state of the command line arguments.
        return input_path, output_path, validation_enabled, enable_scaling_study

    def landbosse_input_dir(self):
        """
        See the get_input_output_paths_from_argv_or_env() function above. This
        function is simply a wrapper around that function to get the input
        path.

        Returns
        -------
        str
            The input directory.
        """

        # The last three elements of the returned tuple are not used here.
        # See get_input_output_paths_from_argv_or_env() docstring for more
        # details.

        input_path, _, _, _ = self.get_input_output_paths_from_argv_or_env()
        return input_path

    def landbosse_output_dir(self):
        """
        See the get_input_output_paths_from_argv_or_env() function above. This
        method gets the base path from there. Then, it checks for a timestamped
        directory that matches the timestamp in this instance. If it finds that
        directory, it simply returns the path to that directory. If it does
        not find that directory, it creates the directory and returns the path
        to the newly created directory.

        Returns
        -------
        str
            The output directory.
        """
        _, output_base_path, _, _ = self.get_input_output_paths_from_argv_or_env()
        output_path = os.path.join(output_base_path, f'landbosse-{self.timestamp}')

        if os.path.exists(output_path) and not os.path.isdir(output_path):
            raise FileExistsError(f'Cannot overwrite {output_path} with LandBOSSE data.')
        elif not os.path.exists(output_path):
            os.mkdir(output_path)
            return output_path
        else:
            return output_path

    def parametric_project_data_output_path(self):
        """
        Returns th path to the project output data folder.

        This folder is to put the parameterized project data sheets generated
        during model runs.

        If the directory does not exist, it is created.

        Returns
        -------
        str
            Path to project data output folder.
        """
        path = os.path.join(self.landbosse_output_dir(), 'calculated_parametric_inputs', 'parametric_project_data')

        if os.path.exists(path) and not os.path.isdir(path):
            raise XlsxOperationException(f'Attempt to write project data to {path} failed. File exists and is not a directory.')

        os.makedirs(path, exist_ok=True)
        return path

    def extended_project_list_path(self):
        """
        This returns the path to which the extended project list, which has all
        the parametric values, should be copied.

        If the folder does not exist yet, this method creates it.

        Returns
        -------
        str
            The absolute path to the destiantion of the extended project
            list.
        """
        path = os.path.join(self.landbosse_output_dir(), 'calculated_parametric_inputs')

        if os.path.exists(path) and not os.path.isdir(path):
            raise XlsxOperationException(f'Attempt to write project data to {path} failed. File exists and is not a directory.')

        os.makedirs(path, exist_ok=True)
        return path

    def copy_input_data(self):
        """
        This copies all input data to the outputs folder. The input data it copies
        are all the data BEFORE they have been modified for parametric runs.
        """
        dst_inputs_copy_path = os.path.join(self.landbosse_output_dir(), 'inputs')
        os.makedirs(dst_inputs_copy_path, exist_ok=True)

        src_project_list_xlsx = os.path.join(self.landbosse_input_dir(), 'project_list.xlsx')
        dst_project_list_xlsx = os.path.join(dst_inputs_copy_path, 'project_list.xlsx')

        src_project_data_dir = os.path.join(self.landbosse_input_dir(), 'project_data')
        dst_project_data_dir = os.path.join(dst_inputs_copy_path, 'project_data')

        copy2(src_project_list_xlsx, dst_project_list_xlsx)
        copytree(src_project_data_dir, dst_project_data_dir)

        src_expected_validation_data = os.path.join(self.landbosse_input_dir(),
                                                    'landbosse-expected-validation-data.xlsx')

        dst_expected_validation_data = os.path.join(self.landbosse_output_dir(),
                                                    'landbosse-expected-validation-data.xlsx')

        if os.path.isfile(src_expected_validation_data):
            copy2(src_expected_validation_data, dst_expected_validation_data)



import traceback
import math

import pandas as pd
import numpy as np
import math
from sympy.solvers import solve
from sympy import Symbol

from .WeatherDelay import WeatherDelay as WD
from .CostModule import CostModule


class FoundationCost(CostModule):
    """
    **FoundationCost.py**

    - Created by Annika Eberle and Owen Roberts on Apr. 3, 2018

    - Refactored by Parangat Bhaskar and Alicia Key on June 3, 2019

    \nCalculates the costs of constructing foundations for land-based wind projects *(items in brackets are not yet implemented)*:

    * Get number of turbines
    * Get duration of construction
    * Get daily hours of operation*  # todo: add to process diagram
    * Get season of construction*  # todo: add to process diagram
    * [Get region]
    * Get rotor diameter
    * Get hub height
    * Get turbine rating
    * Get buoyant foundation design flag
    * [Get seismic zone]
    * Get tower technology type
    * Get hourly weather data
    * [Get specific seasonal delays]
    * [Get long-term, site-specific climate data]
    * Get price data
    * Get labor rates
    * Get material prices for steel and concrete
    * [Use region to determine weather data]


    \n\nGiven below is the set of calculations carried out in this module:

    * Calculate the foundation loads using the rotor diameter, hub height, and turbine rating

    * Determine the foundation size based on the foundation loads, buoyant foundation design flag, and type of tower technology

    * Estimate the amount of material needed for foundation construction based on foundation size and number of turbines

    * Estimate the amount of time required to construct foundation based on foundation size, hours of operation, duration of construction, and number of turbines

    * Estimate the additional amount of time for weather delays (currently only assessing wind delays) based on hourly weather data, construction time, hours of operation, and season of construction

    * Estimate the amount of labor required for foundation construction based on foundation size, construction time, and weather delay
        * Calculate number of workers by crew type
        * Calculate man hours by crew type

    * Estimate the amount of equipment needed for foundation construction based on foundation size, construction time, and weather delay
        * Calculate number of equipment by equip type
        * Calculate equipment hours by equip type

    - Calculate the total foundation cost based on amount of equipment, amount of labor, amount of material, and price data.


    **Keys in the input dictionary are the following:**

    depth
        (int) depth of foundation [in m]


    component_data
        (pd.DataFrame) data frame with wind turbine component data

    def __init__(self, input_dict, output_dict, project_name):
        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name


    num_turbines
        (int) total number of turbines in wind farm

    duration_construction
        (int) estimated construction time in months

    num_delays
        (int) Number of delay events

    avg_hours_per_delay
        (float) Average hours per delay event

    std_dev_hours_per_delay
        (float) Standard deviation from average hours per delay event

    delay_speed_m_per_s
        (float) wind speed above which weather delays kick in

    start_delay_hours
        (int)

    mission_time_hours
        (int)

    gust_wind_speed_m_per_s
        (float)

    wind_height_of_interest_m
        (int)

    wind_shear_exponent
        (float)

    season_construct
        list of seasons (like ['spring', 'summer']) for the construction.

    time_construct
        list of time windows for constructions. Use ['normal'] for a
        0800 to 1800 schedule 10 hour schedule. Use ['long'] for an
        overnight 1800 to 2359, 0000 to 0759 overnight schedule. Use
        ['normal', 'long'] for a 24-hour schedule.

    operational_hrs_per_day
        (float)


    material_price
        (pd.DataFrame) dataframe containing foundation cost related material prices

    rsmeans
        (pd.DataFrame) TODO: Formal definition for rsmeans?


    **Keys in the output dictionary are the following:**

    F_dead_kN_per_turbine
        (float) foundation dead load [in kN]

    F_horiz_kN_per_turbine
        (float) total lateral load [kN]

    M_tot_kN_m_per_turbine
        (float) Moment [kN.m]

    Radius_o_m
        (float) foundation radius based on overturning moment [in m]

    Radius_g_m
        (float) foundation radius based on gapping [in m]

    Radius_b_m
        (float) foundation radius based on bearing pressure [in m]

    Radius_m
        (float) largest foundation radius based on all three foundation design criteria: moment, gapping, bearing [in m]

    foundation_volume_concrete_m3_per_turbine
        (float) volume of a round, raft foundation [in m^3]

    steel_mass_short_ton
        (float) short tons of reinforcing steel

    material_needs_per_turbine
        (pd.DataFrame) table containing material needs info for -> Steel - rebar, Concrete 5000 psi, Excavated dirt, Backfill.

    operation_data
        (pd.DataFrame) TODO: What's the best one line definition for this?


    **TODO: Weather delay set of outputs -> ask Alicia for formal definitions of these keys.**

    total_foundation_cost
        (pd.DataFrame) summary of foundation costs (in USD) broken down into 4 main categories:
        1. Equipment Rental
        2. Labor
        3. Materials
        4. Mobilization
    """

    def __init__(self, input_dict, output_dict, project_name):
        """
        Parameters
        ----------
        input_dict : dict
            The input dictionary with key value pairs described in the
            class documentation

        output_dict : dict
            The output dictionary with key value pairs as found on the
            output documentation.
        """

        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name


        #Constants used in FoundationCost class. Making this data private (hidden from outside of this class): #TODO: Change private variables to protected.
        self._kg_per_tonne = 1000
        self._cubicm_per_cubicft = 0.0283168
        self._steel_density = 9490  # kg / m^3
        self._cubicyd_per_cubicm = 1.30795
        self._ton_per_tonne = 0.907185


    def calculate_foundation_load(self, foundation_load_input_data, foundation_load_output_data):
        """

        Function to calculate foundation load.

        Parameters
        -------
        Int Section height m

        Surface area sq (in m^2)

        Coeff drag (installed)

        Lever arm m (in m)

        Multplier drag rotor

        Multiplier tower drag

        Mass tonne



        Returns
        -------
        Dead load [in N] -> F_dead_kN_per_turbine

        Lateral load [in N] -> F_horiz_kN_per_turbine

        Moment [N.m] -> M_tot_kN_m_per_turbine

        Foundation radius based on overturning moment [in m] -> Radius_o_m

        Foundation radius based on slipping [in m] -> Radius_s_m

        Foundation radius based on gapping [in m] -> Radius_g_m

        Foundation radius based on bearing pressure [in m] -> Radius_b_m

        Largest foundation radius based on all three foundation design criteria (moment, gapping, bearing [in m]) -> Radius_m

        Raises
        ------
        ValueError
            Raises a value error if r_bearing is calculated to be a negative value.
        """
        # set exposure constants
        a = 9.5
        z_g = 274.32

        # get section height
        z = foundation_load_input_data['Section height m']

        # get cross-sectional area
        a_f = foundation_load_input_data['Surface area sq m']

        # get coefficient of drag
        c_d = foundation_load_input_data['Coeff drag (installed)']

        # get lever arm
        l = foundation_load_input_data['Lever arm m']

        # get multipliers for tower and rotor
        multiplier_rotor = foundation_load_input_data['Multplier drag rotor']
        multiplier_tower = foundation_load_input_data['Multiplier tower drag']


        # calculate wind pressure
        k_z = 2.01 * (z / z_g) ** (2 / a)  # exposure factor
        k_d = 0.95  # wind directionality factor
        k_zt = 1  # topographic factor
        v = foundation_load_input_data['gust_velocity_m_per_s']
        wind_pressure = 0.613 * k_z * k_zt * k_d * v ** 2

        # calculate wind loads on each tower component
        g = 0.85  # gust factor
        c_f = 0.6  # coefficient of force
        f_t = (wind_pressure * g * c_f * a_f) * multiplier_tower

        # calculate drag rotor
        rho = 1.225  # air density in kg/m^3
        f_r = (0.5 * rho * c_d * a_f * v ** 2) * multiplier_rotor

        f = (f_t + f_r)

        # calculate dead load in N
        g = 9.8  # m / s ^ 2
        f_dead = sum(foundation_load_input_data['Mass tonne']) * g * self._kg_per_tonne / 1.15  # scaling factor to adjust dead load for uplift

        # calculate moment from each component at base of tower
        m_overturn = f * l

        # get total lateral load (N) and moment (N * m)
        f_lat = f.sum()  # todo: add f_lat (drag force) to output csv
        m_overturn = m_overturn.sum()

        # compare to moment from rated thrust
        rated_thrust = foundation_load_input_data['rated_thrust_N']
        m_thrust = rated_thrust * max(l)
        m_tot = max(m_thrust, m_overturn)

        # compare lateral load to rated thrust
        f_horiz = max(f_lat, rated_thrust)

        # calculate foundation radius based on overturning moment
        vol_fraction_fill = 0.55
        vol_fraction_concrete = 1 - vol_fraction_fill
        safety_overturn = 1.5
        unit_weight_fill = 17.3e3  # in N / m^3
        unit_weight_concrete = 23.6e3  # in N / m^3
        bearing_pressure = foundation_load_input_data['bearing_pressure_n_m2']
        p = [(np.pi * foundation_load_input_data['depth'] * (vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete)), 0, f_dead, - (safety_overturn * (m_tot + f_horiz * foundation_load_input_data['depth']))]
        r_overturn = np.roots(p)
        r_overturn = np.real(r_overturn[np.isreal(r_overturn)])[0]

        # calculate foundation radius based on slipping
        safety_slipping = 1.5
        friction_angle_soil = 25
        tangent_slip_angle = math.tan((friction_angle_soil * math.pi) / 180)
        slipping_force_with_sf = (safety_slipping * f_lat)
        # first check if slipping is already satisfied by dead weight
        if slipping_force_with_sf < (f_dead * tangent_slip_angle):
            r_slipping = 0
        else:
            # Calculate foundation radius based on slipping:
            r_slipping = (((slipping_force_with_sf / tangent_slip_angle) - f_dead) /
                          ((vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete) * math.pi * foundation_load_input_data['depth'])) ** 0.5

        r_test_gapping = max(r_overturn, r_slipping)

        # calculate foundation radius based on gapping
        # check if gapping constrain is already satisfied - r / 3 < e
        foundation_vol = np.pi * r_test_gapping ** 2 * foundation_load_input_data['depth']
        v_1 = (foundation_vol * (
                    vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete) + f_dead)
        e = m_tot / v_1
        if (r_test_gapping / 3) < e:
            r_gapping = 0
        else:
            r_g = Symbol('r_g', real=True, positive=True)
            foundation_vol = np.pi * r_g ** 2 * foundation_load_input_data['depth']
            v_1 = (foundation_vol * (vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete) + f_dead)
            e = m_tot / v_1
            r_gapping = solve(e * 3 - r_g, r_g)
            if len(r_gapping) > 0:
                r_gapping = max(r_gapping)
            else:
                r_gapping = 0

        r_test_bearing = max(r_test_gapping, r_gapping)

        # calculate foundation radius based on bearing pressure

        # Restrict r_b to only real numbers. Positive solutions for r_b are
        # selected below
        r_b = Symbol('r_b', real=True)

        foundation_vol = np.pi * r_test_bearing ** 2 * foundation_load_input_data['depth']
        v_1 = (foundation_vol * (vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete) + f_dead)
        e = m_tot / v_1
        a_eff = v_1 / bearing_pressure
        r_bearing = solve(2 * (r_b ** 2 - e * (r_b ** 2 - e ** 2) ** 0.5) - a_eff, r_b)

        # Select only positive solutions to r_b. This is selected by max(). If there are
        # not positive solutions to r_b, that means something is wrong with the foundation
        # parameters. In that case, generate a warning below.

        if len(r_bearing) > 0:
            r_bearing = max(r_bearing)
        else:
            r_bearing = 0

        if r_bearing < 0:
            raise ValueError(f'Warning {self.project_name} calculate_foundation_load r_bearing is negative, r_bearing={r_bearing}')

        # pick the largest foundation radius based on all 4 foundation design criteria: moment, gapping, bearing, slipping
        r_choosen = max(r_bearing, r_overturn, r_slipping, r_gapping)

        foundation_load_output_data['F_dead_kN_per_turbine']    =   f_dead / 1e3
        foundation_load_output_data['F_horiz_kN_per_turbine']   =   f_lat / 1e3
        foundation_load_output_data['M_tot_kN_m_per_turbine']   =   m_tot / 1e3
        foundation_load_output_data['Radius_o_m']   =   r_overturn
        foundation_load_output_data['Radius_s_m']   =   r_slipping
        foundation_load_output_data['Radius_g_m']   =   r_gapping
        foundation_load_output_data['Radius_b_m']   =   r_bearing
        foundation_load_output_data['Radius_m']     =   r_choosen

        return foundation_load_output_data


    def determine_foundation_size(self, foundation_size_input_data, foundation_size_output_data):
        """
        Function to calculate the volume of a round, raft foundation. Assumes foundation made of concrete with 1 m thickness.

        Parameters
        -------
        Largest foundation radius based on all three foundation design criteria: moment, gapping, bearing [in m] -> Radius_m [in m]

        depth of foundation [in m] -> depth


        Returns
        -------
        Foundation volume [in m^3] -> foundation_volume_concrete_m3_per_turbine

        """
        #TODO: still updating/fine-tuning foundation size equations for small DW (Parangat - Feb 27, 2020)
        r = float(foundation_size_output_data['Radius_m'])
        if foundation_size_input_data['turbine_rating_MW'] < 0.1:
            foundation_size_output_data['excavated_volume_m3'] = r * r * foundation_size_input_data['depth']
            foundation_size_output_data['foundation_volume_concrete_m3_per_turbine'] = foundation_size_output_data['excavated_volume_m3'] * 0.45
        else:
            foundation_size_output_data['excavated_volume_m3'] = np.pi * (r + 0.5) ** 2 * foundation_size_input_data['depth']

            # only compute the portion of the foundation that is composed of concrete (45% concrete; other portion is
            # backfill); TODO: Add to sphinx -> (volume excavated = pi*(r_pick + .5m)^2 this assumes vertical sides which
            #  does not reflect reality as OSHA requires benched sides over 3’)
            foundation_size_output_data['foundation_volume_concrete_m3_per_turbine'] = np.pi * r ** 2 * \
                                                                                       foundation_size_input_data['depth'] * 0.45

        return foundation_size_output_data


    def estimate_material_needs_per_turbine(self, material_needs_per_turbine_input_data, material_needs_per_turbine_output_data):
        """
        Function to estimate amount of material based on foundation size and number of turbines.


        Parameters
        -------
        Foundation concrete volume [in m^3] -> foundation_volume_concrete_m3_per_turbine


        Returns
        -------

        (Returns pd.DataFrame) material_needs_per_turbine


        """

        steel_mass_short_ton_per_turbine = (material_needs_per_turbine_output_data['foundation_volume_concrete_m3_per_turbine'] * 0.012 * self._steel_density / self._kg_per_tonne)
        concrete_volume_cubic_yards_per_turbine = material_needs_per_turbine_output_data['foundation_volume_concrete_m3_per_turbine'] * 0.985 * self._cubicyd_per_cubicm

        #Assign values to output dictionary:
        material_needs_per_turbine_output_data['material_needs_per_turbine'] = pd.DataFrame([['Steel - rebar', steel_mass_short_ton_per_turbine, 'ton (short)'],
                                                                     ['Concrete 5000 psi', concrete_volume_cubic_yards_per_turbine, 'cubic yards'],
                                                                     ['Excavated dirt', material_needs_per_turbine_output_data['excavated_volume_m3'] * self._cubicyd_per_cubicm, 'cubic_yards'],
                                                                     ['Backfill', material_needs_per_turbine_output_data['excavated_volume_m3'] * self._cubicyd_per_cubicm, 'cubic_yards']],
                                                                    columns=['Material type ID', 'Quantity of material', 'Units'])

        material_needs_per_turbine_output_data['steel_mass_short_ton_per_turbine'] = steel_mass_short_ton_per_turbine
        return material_needs_per_turbine_output_data['material_needs_per_turbine']


    def estimate_construction_time(self, construction_time_input_data, construction_time_output_data):
        """
        Function to estimate construction time on per turbine basis. TODO: What's a better definition of this function. It's task is to return a pd.DataFrame (operation_data).

        Parameters
        -------
        duration_construction

        pd.DataFrame
            rsmeans

        pd.DataFrame
            material_needs_per_turbine



        Returns
        -------

        (pd.DataFrame) operation_data

        """

        foundation_construction_time = construction_time_input_data['construct_duration'] * 1 / 3
        #throughput_operations = construction_time_input_data['throughput_operations']
        throughput_operations = construction_time_input_data['rsmeans']
        material_needs_per_turbine = construction_time_output_data['material_needs_per_turbine']
        quantity_materials_entire_farm = material_needs_per_turbine['Quantity of material'] * construction_time_input_data['num_turbines']


        #Calculations for estimate construction time will be on entire wind farm basis:
        construction_time_output_data['material_needs_entire_farm'] = material_needs_per_turbine.copy()
        material_needs_entire_farm = construction_time_output_data['material_needs_entire_farm']
        material_needs_entire_farm['Quantity of material'] = quantity_materials_entire_farm
        if construction_time_input_data['turbine_rating_MW'] <= 0.1:
            operation_data = throughput_operations.where(
                throughput_operations['Module'] == 'Small DW Foundations').dropna(
                thresh=4)
        else:
            operation_data = throughput_operations.where(throughput_operations['Module'] == 'Foundations').dropna(
                thresh=4)

        #operation data for entire wind farm:
        operation_data = pd.merge(material_needs_entire_farm, operation_data, on=['Material type ID'], how='outer')
        operation_data['Number of days'] = operation_data['Quantity of material'] / operation_data['Daily output']
        operation_data['Number of crews'] = np.ceil((operation_data['Number of days'] / 30) / foundation_construction_time)

        alpha = operation_data[operation_data['Type of cost'] == 'Labor']
        operation_data_id_days_crews_workers = alpha[['Operation ID', 'Number of days', 'Number of crews', 'Number of workers']]

        # if more than one crew needed to complete within construction duration then assume that all construction happens
        # within that window and use that timeframe for weather delays; if not, use the number of days calculated
        operation_data['time_construct_bool'] = operation_data['Number of days'] > foundation_construction_time * 30
        boolean_dictionary = {True: foundation_construction_time * 30, False: np.NAN}
        operation_data['time_construct_bool'] = operation_data['time_construct_bool'].map(boolean_dictionary)
        operation_data['Time construct days'] = operation_data[['time_construct_bool', 'Number of days']].min(axis=1)
        num_days = operation_data['Time construct days'].max()

        construction_time_output_data['operation_data_id_days_crews_workers'] = operation_data_id_days_crews_workers
        construction_time_output_data['operation_data_entire_farm'] = operation_data

        # pull out management data #TODO: Add this cost to Labor cost next
        if construction_time_input_data['turbine_rating_MW'] > 0.1:
            crew_cost = self.input_dict['crew_cost']
            crew = self.input_dict['crew'][self.input_dict['crew']['Crew type ID'].str.contains('M0')]
            management_crew = pd.merge(crew_cost, crew, on=['Labor type ID'])
            management_crew = management_crew.assign(per_diem_total=management_crew['Per diem USD per day'] * management_crew['Number of workers'] * num_days)
            management_crew = management_crew.assign(hourly_costs_total=management_crew['Hourly rate USD per hour'] * self.input_dict['hour_day'][self.input_dict['time_construct']] * num_days)
            management_crew = management_crew.assign(total_crew_cost_before_wind_delay=management_crew['per_diem_total'] + management_crew['hourly_costs_total'])
            self.output_dict['management_crew'] = management_crew
            self.output_dict['managament_crew_cost_before_wind_delay'] = management_crew['total_crew_cost_before_wind_delay'].sum()
        else:
            self.output_dict['managament_crew_cost_before_wind_delay'] = 0

        return construction_time_output_data['operation_data_entire_farm']


    def calculate_weather_delay(self, weather_delay_input_data, weather_delay_output_data):
        """
        Function to calculate wind delay for foundations.

        Keys in weather_delay_input_data
        --------------------------------
        weather_window

        duration_construction

        start_delay

        critical_wind_speed

        operational_hrs_per_day

        height_interest

        wind_shear_exponent
        """

        # construct WeatherDelay module
        WD(weather_delay_input_data, weather_delay_output_data)

        # compute weather delay
        wind_delay = pd.DataFrame(weather_delay_output_data['wind_delays'])

        # if greater than 4 hour delay, then shut down for full day (10 hours)
        wind_delay[(wind_delay > 4)] = 10
        weather_delay_output_data['wind_delay_time'] = float(wind_delay.sum())

        return weather_delay_output_data


    def calculate_costs(self, calculate_costs_input_dict, calculate_costs_output_dict):
        """
        Function to calculate the total foundation cost.

        Keys in input dictionary
        ------------------------
        pd.DataFrame
            material_needs_per_turbine

        pd.DataFrame
            material_price

        pd.DataFrame
            operation_data

        wind_delay_time

        operational_hrs_per_day

        wind_multiplier

        pd.DataFrame
            rsmeans


        Returns
        -------

        (pd.DataFrame) total_foundation_cost


        """


        material_vol_entire_farm = calculate_costs_output_dict['material_needs_entire_farm']
        material_price = calculate_costs_input_dict['material_price']


        material_data_entire_farm = pd.merge(material_vol_entire_farm, material_price, on=['Material type ID'])
        material_data_entire_farm['Cost USD'] = material_data_entire_farm['Quantity of material'] * pd.to_numeric(material_data_entire_farm['Material price USD per unit'])     # material data on a total wind farm basis


        operation_data = calculate_costs_output_dict['operation_data_entire_farm']

        wind_delay = calculate_costs_output_dict['wind_delay_time']

        wind_delay_fraction = (wind_delay / calculate_costs_input_dict['operational_hrs_per_day']) / operation_data['Time construct days'].max(skipna=True)
        # check if wind_delay_fraction is greater than 1, which would mean weather delays are longer than they can possibily be for the input data
        if wind_delay_fraction > 1:
            raise ValueError('{}: Error: Wind delay greater than 100%'.format(type(self).__name__))
        wind_multiplier = 1 / (1 - wind_delay_fraction)
        calculate_costs_output_dict['wind_multiplier'] = wind_multiplier

        rsmeans = calculate_costs_input_dict['rsmeans']
        if calculate_costs_input_dict['turbine_rating_MW'] > 0.1:
            rsmeans = rsmeans.where(rsmeans['Module'] == 'Foundations').dropna(thresh=4)
        else:
            rsmeans = rsmeans.where(rsmeans['Module'] == 'Small DW Foundations').dropna(thresh=4)

        labor_equip_data = pd.merge(material_vol_entire_farm, rsmeans, on=['Material type ID'])

        # Create foundation cost dataframe
        foundation_cost = pd.DataFrame(columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        # Calculate per diem
        per_diem = operation_data['Number of workers'] * operation_data['Number of crews'] * (operation_data['Time construct days'] +
                                                                                              np.ceil(operation_data['Time construct days'] / 7)) * calculate_costs_input_dict['rsmeans_per_diem']
        where_are_na_ns = np.isnan(per_diem)
        per_diem[where_are_na_ns] = 0
        labor_equip_data['Cost USD'] = (labor_equip_data['Quantity of material'] * labor_equip_data['Rate USD per unit'] * calculate_costs_input_dict['overtime_multiplier'] + per_diem + calculate_costs_output_dict['managament_crew_cost_before_wind_delay']) * wind_multiplier
        self.output_dict['labor_equip_data'] = labor_equip_data

        # EQUIPMENT COST
        # Create equipment costs row to be appended to foundation_cost
        equipment_dataframe = labor_equip_data[labor_equip_data['Type of cost'].str.match('Equipment rental')]

        equipment_cost_usd_without_delay = (
                equipment_dataframe['Quantity of material'] * equipment_dataframe['Rate USD per unit'] *
                calculate_costs_input_dict['overtime_multiplier'] + per_diem)
        equipment_cost_usd_with_weather_delays = equipment_cost_usd_without_delay.sum() * wind_multiplier
        equipment_costs = pd.DataFrame([['Equipment rental', equipment_cost_usd_with_weather_delays, 'Foundation']],
                                       columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        # LABOR COST
        # Create labor costs row to be appended to foundation_cost
        labor_dataframe = labor_equip_data[labor_equip_data['Type of cost'].str.match('Labor')]
        labor_cost_usd_without_management= (labor_dataframe['Quantity of material'] * labor_dataframe['Rate USD per unit'] * calculate_costs_input_dict['overtime_multiplier'] + per_diem )
        labor_cost_usd_with_management = labor_cost_usd_without_management.sum() + calculate_costs_output_dict['managament_crew_cost_before_wind_delay']
        labor_cost_usd_with_management_plus_weather_delays = labor_cost_usd_with_management * wind_multiplier
        labor_costs = pd.DataFrame([['Labor', labor_cost_usd_with_management_plus_weather_delays, 'Foundation']],
                                   columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        # MATERIAL COST
        material_cost_dataframe = pd.DataFrame(columns=['Operation ID', 'Type of cost', 'Cost USD'])
        material_cost_dataframe['Operation ID'] = material_data_entire_farm['Material type ID']
        material_cost_dataframe['Type of cost'] = 'Materials'
        material_cost_dataframe['Cost USD'] = material_data_entire_farm['Cost USD']
        material_costs_sum = material_cost_dataframe['Cost USD'].sum()
        material_costs = pd.DataFrame([['Materials', material_costs_sum, 'Foundation']],
                                      columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        # Append all cost items to foundation_cost
        foundation_cost = foundation_cost.append(equipment_costs)
        foundation_cost = foundation_cost.append(labor_costs)
        foundation_cost = foundation_cost.append(material_costs)

        # Calculate mobilization cost as percentage of total foundation cost and add to foundation_cost
        # Assumed 5% of total foundation cost and add to foundation_cost for utility scale plant
        # A function of turbine size for distributed wind (< 10 turbines)
        if calculate_costs_input_dict['num_turbines'] > 10:
            mobilization_cost = foundation_cost['Cost USD'].sum() * 0.05
        else:
            if calculate_costs_input_dict['turbine_rating_MW'] < 0.1:
                # Zero since mobilization cost of equipment is included in the equipment rental cost
                mobilization_cost = 0
            else:
                # There is mobilization cost for 0-10 turbines 100+ kW in rating.
                num_turbines = calculate_costs_input_dict['num_turbines']
                rating = calculate_costs_input_dict['turbine_rating_MW']
                mobilization_multipler = self.mobilization_cost_multiplier(rating)
                mobilization_cost = foundation_cost['Cost USD'].sum() / num_turbines * mobilization_multipler

        mob_cost = pd.DataFrame([['Mobilization', mobilization_cost, 'Foundation']], columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        foundation_cost = foundation_cost.append(mob_cost)

        # For LandBOSSE API, cost breakdown by type stored as floating point values:
        calculate_costs_output_dict['foundation_equipment_rental_usd'] = equipment_cost_usd_with_weather_delays
        calculate_costs_output_dict['foundation_labor_usd'] = labor_cost_usd_with_management_plus_weather_delays
        calculate_costs_output_dict['foundation_material_usd'] = material_costs_sum
        calculate_costs_output_dict['foundation_mobilization_usd'] = mobilization_cost


        # todo: we add a separate tab in the output file for costs (all costs will be the same format but it's a different format than other data)
        # columns in cost tab would include project_id, module, operation_id, type_of_cost, total_or_per_turbine, cost_usd
        # an example row for this type of output would be "project1, FoundationCost, Rebar installation, Labor, total, 2.371127e+06"
        # total_foundation_cost = foundation_cost.groupby(by=['Type of cost']).sum().reset_index()
        # total_foundation_cost['Phase of construction'] = 'Foundations'
        # total_cost_summed_foundation = total_foundation_cost.sum(numeric_only=True)[0] # todo: add total_cost_summed_foundation to output dict

        total_foundation_cost = foundation_cost
        calculate_costs_output_dict['total_foundation_cost'] = total_foundation_cost
        calculate_costs_output_dict['summed_foundation_cost'] = total_foundation_cost['Cost USD'].sum()   #for landbosse_api

        self.output_dict['labor_equip_data'] = labor_equip_data

        return total_foundation_cost

    def outputs_for_detailed_tab(self, input_dict, output_dict):
        """
        Creates a list of dictionaries which can be used on their own or
        used to make a dataframe.

        Must be called after self.run_module()

        Returns
        -------
        list(dict)
            A list of dicts, with each dict representing a row of the data.
        """

        # Note that some values are cast with float() so that XlsxWriter
        # (the library depended on by XlsxGenerator) can output them as
        # numbers. XlsxWriter, interestingly, cannot handle np.float32()
        # types.

        result = []
        module = type(self).__name__
        result.append({
            'unit': '',
            'type': 'variable',
            'variable_df_key_col_name': 'wind_multiplier',
            'value': float(self.output_dict['wind_multiplier'])
        })
        result.append({
            'unit': 'kN',
            'type': 'variable',
            'variable_df_key_col_name': 'F_dead',
            'value': float(self.output_dict['F_dead_kN_per_turbine'])
        })
        result.append({
            'unit': 'kN',
            'type': 'variable',
            'variable_df_key_col_name': 'F_horiz',
            'value': float(self.output_dict['F_horiz_kN_per_turbine'])
        })
        result.append({
            'unit': 'kN_m',
            'type': 'variable',
            'variable_df_key_col_name': 'M_tot_kN',
            'value': float(self.output_dict['M_tot_kN_m_per_turbine'])
        })
        result.append({
            'unit': 'm',
            'type': 'variable',
            'variable_df_key_col_name': 'Radius_o',
            'value': float(self.output_dict['Radius_o_m'])
        })
        result.append({
            'unit': 'm',
            'type': 'variable',
            'variable_df_key_col_name': 'Radius_g',
            'value': float(self.output_dict['Radius_g_m'])
        })
        result.append({
            'unit': 'm',
            'type': 'variable',
            'variable_df_key_col_name': 'Radius_b',
            'value': float(self.output_dict['Radius_b_m'])
        })
        result.append({
            'unit': 'm',
            'type': 'variable',
            'variable_df_key_col_name': 'Radius',
            'value': float(self.output_dict['Radius_m'])
        })
        result.append({
            'unit': 'm^3',
            'type': 'variable',
            'variable_df_key_col_name': 'foundation_volume_concrete_m3_per_turbine',
            'value': float(self.output_dict['foundation_volume_concrete_m3_per_turbine'])
        })
        result.append({
            'unit': 'short_ton',
            'type': 'variable',
            'variable_df_key_col_name': 'steel_mass_short_ton_per_turbine',
            'value': self.output_dict['steel_mass_short_ton_per_turbine']
        })
        # foundation_volume_concrete_m3_per_turbine
        result.append({
            'unit': 'm^3',
            'type': 'variable',
            'variable_df_key_col_name': 'foundation_volume_concrete_m3_per_turbine',
            'value': self.output_dict['foundation_volume_concrete_m3_per_turbine']
        })

        for row in self.output_dict['operation_data_id_days_crews_workers'].itertuples():
            dashed_row = '{}-{}-{}-{}'.format(row[1], math.ceil(row[2]), row[3], row[4])
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'operation_data: Operation ID-Number of days-Number of crews-Number of workers',
                'value': dashed_row
            })

        for row in self.output_dict['material_needs_per_turbine'].itertuples():
            # This must be formatted in Python
            dashed_row = '{}-{}-{:.2e}'.format(row[0], row[1], row[2])
            result.append({
                'unit': row[3],
                'type': 'dataframe',
                'variable_df_key_col_name': 'material_needs_per_turbine: {}'.format('-'.join(self.output_dict['material_needs_per_turbine'].columns[:-1])),
                'value': dashed_row
            })

        for row in self.output_dict['total_foundation_cost'].itertuples():
            dashed_row = '{} <--> {} <--> {}'.format(row[1], row[3], math.ceil(row[2]))
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'Type of Cost <--> Phase of Construction <--> Cost in USD ',
                'value': dashed_row,
                'last_number': row[2]
            })

        for _dict in result:
            _dict['project_id_with_serial'] = self.project_name
            _dict['module'] = module

        self.output_dict['foundation_cost_csv'] = result
        return result

    def outputs_for_module_type_operation(self, input_dict, output_dict):
        result = []
        module = type(self).__name__

        costs_by_module_type_operation = self.output_dict['labor_equip_data']
        for _, row in costs_by_module_type_operation.iterrows():
            _dict = dict()
            row = row.to_dict()
            _dict['operation_id'] = row['Operation ID']
            _dict['type_of_cost'] = row['Type of cost']
            _dict['cost'] = row['Cost USD']
            result.append(_dict)

        for _dict in result:
            _dict['project_id_with_serial'] = self.project_name
            _dict['module'] = module
            _dict['total_or_turbine'] = 'total'

        self.output_dict['foundation_module_type_operation'] = result
        return result

    def run_module(self):
        """
        Runs the FoundationCost module and populates the IO dictionaries with calculated values.

        Parameters
        ----------
        <None>

        Returns
        -------
        tuple
            First element of tuple contains a 0 or 1. 0 means no errors happened and
            1 means an error happened and the module failed to run. The second element
            either returns a 0 if the module ran successfully, or it returns the error
            raised that caused the failure.

        """
        try:
            self.calculate_foundation_load(self.input_dict, self.output_dict)  # Returns foundation load
            self.determine_foundation_size(self.input_dict, self.output_dict)  # Returns foundation volume
            self.estimate_material_needs_per_turbine(self.input_dict, self.output_dict)  # Returns material volume
            operation_data = self.estimate_construction_time(self.input_dict, self.output_dict)  # Estimates construction time

            # pull only global inputs for weather delay from input_dict
            weather_data_keys = ('wind_shear_exponent',
                                 'weather_window')

            # specify foundation-specific weather delay inputs
            self.weather_input_dict = dict(
                [(i, self.input_dict[i]) for i in self.input_dict if i in set(weather_data_keys)])
            self.weather_input_dict[
                'start_delay_hours'] = 0  # assume zero start for when foundation construction begins (start at beginning of construction time)
            self.weather_input_dict[
                'critical_wind_speed_m_per_s'] = self.input_dict['critical_speed_non_erection_wind_delays_m_per_s']
            self.weather_input_dict[
                'wind_height_of_interest_m'] = self.input_dict['critical_height_non_erection_wind_delays_m']

            # duration_construction is in units of days
            # duration_construction_months is in units of months
            days_per_month = 30
            duration_construction = operation_data['Time construct days'].max(skipna=True)
            duration_construction_months = duration_construction / days_per_month
            self.output_dict['foundation_construction_months'] = duration_construction_months

            # compute and specify weather delay mission time for roads
            operational_hrs_per_day = self.input_dict['hour_day'][self.input_dict['time_construct']]
            mission_time_hrs = duration_construction * operational_hrs_per_day
            self.weather_input_dict['mission_time_hours'] = mission_time_hrs
            self.input_dict['operational_hrs_per_day'] = operational_hrs_per_day

            self.calculate_weather_delay(self.weather_input_dict, self.output_dict)
            self.calculate_costs(self.input_dict, self.output_dict)
            self.outputs_for_detailed_tab(self.input_dict, self.output_dict)
            # self.output_dict['labor_equip_data']
            # self.output_dict['foundation_module_type_operation'] = self.outputs_for_module_type_operation(self.input_dict, self.output_dict)
            self.output_dict['foundation_module_type_operation'] = self.outputs_for_costs_by_module_type_operation(
                input_df=self.output_dict['total_foundation_cost'],
                project_id=self.project_name,
                total_or_turbine=True
            )
            return 0, 0   # module ran successfully
        except Exception as error:
            traceback.print_exc()
            print(f"Fail {self.project_name} FoundationCost")
            self.input_dict['error']['FoundationCost'] = error
            return 1, error    # module did not run successfully



import os
import pandas as pd
import numpy as np
from ..excelio import XlsxReader
from ..excelio import XlsxReader
from ..excelio.WeatherWindowCSVReader import read_weather_window
from ..excelio.XlsxDataframeCache import XlsxDataframeCache
from ..model import Manager
from datetime import datetime, timedelta
from hybridbosse.LandBOSSE.landbosse.landbosse_api.turbine_scaling import nacelle_mass, \
                                            edit_nacelle_info, \
                                            hub_mass, \
                                            edit_hub_info, \
                                            blade_mass_ton, \
                                            edit_blade_info, \
                                            tower_mass, \
                                            tower_specs, \
                                            edit_tower_sections


def run_landbosse(input_dict):
    """
    Call this function - run_landbosse() - to run LandBOSSE.
    This method calls the read_data() method to read the 2 excel input files that
    are packaged with LandBOSSE, and creates a master input dictionary from it.

    The 2 excel input files packaged with LandBOSSE are:
        1. project_list.xlxs : This file defines the high level project inputs
        that define a unique wind farm project
        2. project_data_defaults.xlsx

    Parameters
        ----------
       input_dict : Python Dictionary
           This input dictionary is a required argument for this function. It
           consists of 13 required key:value pairs, and 1 optional key:value
           pair ('weather_file_path'). The key:value pairs are as follows:


                                KEY               | REQUIRED KEY? (Y/N)   | SUGGESTED SAMPLE DEFAULT VALUE  | DESCRIPTION                                               |
                ----------------------------------|-----------------------|---------------------------------|-----------------------------------------------------------|
                'interconnect_voltage_kV'         |           Y           |         137                     | Grid interconnection voltage (kV)                         |
                ---------------------------------------------------------------------------------------------------------------------------------------------------------
                'distance_to_interconnect_mi'     |           Y           |         10                      | Distance to interconnection (miles)                       |
                --------------------------------------------------------------------------------------------------------------------------------------------------------
                'num_turbines'                    |           Y           |         100                     | Total number of turbines in project                       |
                --------------------------------------------------------------------------------------------------------------------------------------------------------|
                'turbine_spacing_rotor_diameters  |           Y           |         4                       | Turbine spacing (in multiples of turbine rotor diameter)  |
                --------------------------------------------------------------------------------------------------------------------------------------------------------|
                'row_spacing_rotor_diameters'     |           Y           |         10                      | Row spacing (in multiples of turbine rotor diameter)      |
                ------------------------------------------------------------------------------------------------------------------------                                |
                'turbine_rating_MW'               |           Y           |         1.5                     | Turbine rating (MW)                                       |
                --------------------------------------------------------------------------------------------------------------------------------------------------------|
                'rotor_diameter_m'                |           Y           |         77                      | Turbine rotor diameter (meters)                           |
                --------------------------------------------------------------------------------------------------------------------------------------------------------|
                'hub_height_meters'               |           Y           |         80                      | Turbine hub height (meters)                               |
                --------------------------------------------------------------------------------------------------------------------------------------------------------|
                'wind_shear_exponent'             |           Y           |         0.2                     | Wind shear exponent                                       |
                ------------------------------------------------------------------------------------------------------------------------                                |
                'depth'                           |           Y           |         2.36                    | Turbine foundation depth (meters)                         |
                ------------------------------------------------------------------------------------------------------------------------                                |
                'rated_thrust_N'                  |           Y           |         589,000                 | Turbine rated thrust (Newtons)                            |
                ------------------------------------------------------------------------------------------------------------------------                                |
                'labor_cost_multiplier'           |           Y           |         1                       | Labor cost mutliplier                                     |
                ------------------------------------------------------------------------------------------------------------------------                                |
                'gust_velocity_m_per_s'           |           Y           |         59.50                   | 50-year max gust velocity (m/s)                           |
                --------------------------------------------------------------------------------------------------------------------------------------------------------|
                'weather_file_path'               |           N           | <OPTIONAL>Provide an            |   This is the                                             |
                                                  |                       |  absolute path to a .srw        |   weather resurce                                         |
                                                  |                       |  hourly weather file            |   that will help                                          |
                                                  |                       |  (8760 rows).                   |   determine whether                                       |
                                                  |                       |                                 |   weather based delays                                    |
                                                  |                       |  Else, default weather          |   in project construction                                 |
                                                  |                       |  file packaged with             |                                                           |
                                                  |                       |  LandBOSSE will be used.        |                                                           |
                                                  |                       |                                 |                                                           |
                                                  |                       |                                 |                                                           |
                                                  |                       |                                 |                                                           |
                --------------------------------------------------------------------------------------------------------------------------------------------------------|

        Returns
        -------
        output_dict : Python Dictionary
            Output dictionary container with balance-of-station costs calculated by the model. Following outputs are
            returned:

                        OUTPUT                                  |               KEY                    |
            ----------------------------------------------------|--------------------------------------|
            Total Management Cost                               |   total_management_cost               |
            --------------------------------------------------------------------------------------------
            Total Insurance Cost (USD)                          |   insurance_usd                       |
            --------------------------------------------------------------------------------------------
            Total Construction Permitting Cost (USD)            |   construction_permitting_usd         |
            --------------------------------------------------------------------------------------------
            Total Project Management Cost (USD)                 |   project_management_usd              |
            --------------------------------------------------------------------------------------------
            Total Bonding Cost (USD)                            |   bonding_usd                         |
            --------------------------------------------------------------------------------------------
            Total Markup Contingency Cost (USD)                 |   markup_contingency_usd              |
            --------------------------------------------------------------------------------------------
            Total Engineering Cost (USD)                        |   engineering_usd                     |
            --------------------------------------------------------------------------------------------
            Total Site Facility Cost (USD)                      |   site_facility_usd                   |
            --------------------------------------------------------------------------------------------
            Total Management Cost (USD)                         |   total_management_cost               |
            --------------------------------------------------------------------------------------------
            Total Development Cost (USD)                        |   summed_development_cost             |
            --------------------------------------------------------------------------------------------
            Total Development Labor Cost (USD)                  |   development_labor_cost_usd          |
            --------------------------------------------------------------------------------------------
            Total Site Preparation Cost (USD)                   |   summed_sitepreparation_cost         |
            --------------------------------------------------------------------------------------------
            Total Site Preparation Equipment Rental Cost (USD)  |   sitepreparation_equipment_rental_usd |
            --------------------------------------------------------------------------------------------
            Total Site Preparation Labor Cost (USD)             |   sitepreparation_labor_usd           |
            --------------------------------------------------------------------------------------------
            Total Site Preparation Material Cost (USD)          |   sitepreparation_material_usd        |
            --------------------------------------------------------------------------------------------
            Total Site Preparation Mobilization Cost (USD)      |   sitepreparation_mobilization_usd    |
            --------------------------------------------------------------------------------------------
            Total Foundation Cost (USD)                         |   summed_foundation_cost              |
            --------------------------------------------------------------------------------------------
            Total Foundation Equipment Rental Cost (USD)        |   foundation_equipment_rental_usd     |
            --------------------------------------------------------------------------------------------
            Total Foundation Labor Cost (USD)                   |   foundation_labor_usd                |
            --------------------------------------------------------------------------------------------
            Total Foundation Material Cost (USD)                |   foundation_material_usd             |
            --------------------------------------------------------------------------------------------
            Total Foundation Mobilization Cost (USD)            |   foundation_mobilization_usd         |
            --------------------------------------------------------------------------------------------
            Total Tower Erection Cost (USD)                     |   total_cost_summed_erection          |
            --------------------------------------------------------------------------------------------
            Total Tower Erection Equipment Rental Cos (USD)     |   erection_equipment_rental_usd       |
            --------------------------------------------------------------------------------------------
            Total Tower Erection Labor Cost (USD)               |   erection_labor_usd                  |
            --------------------------------------------------------------------------------------------
            Total Tower Erection Material Cost (USD)            |   erection_material_usd               |
            --------------------------------------------------------------------------------------------
            Total Tower Erection - Other Cost (USD)             |   erection_other_usd                  |
            --------------------------------------------------------------------------------------------
            Total Tower Erection Mobilization Cost (USD)        |   erection_mobilization_usd           |
            --------------------------------------------------------------------------------------------
            Total Tower Erection Fuel Cost (USD)                |   erection_fuel_usd                   |
            --------------------------------------------------------------------------------------------
            Total Transmission & Distribution Cost (USD)        |   trans_dist_usd                      |
            --------------------------------------------------------------------------------------------
            Total Collection Cost                               |   summed_collection_cost              |
            --------------------------------------------------------------------------------------------
            Total Collection System Equipment Rental Cost (USD) |   collection_equipment_rental_usd     |
            --------------------------------------------------------------------------------------------
            Total Collection System Labor Cost (USD)            |   collection_labor_usd                |
            --------------------------------------------------------------------------------------------
            Total Collection System Material Cost (USD)         |   collection_material_usd             |
            --------------------------------------------------------------------------------------------
            Total Collection System Mobilization Cost (USD)     |   collection_mobilization_usd         |
            --------------------------------------------------------------------------------------------
            Total Substation Cost (USD)                         |   summed_substation_cost              |
            --------------------------------------------------------------------------------------------

    """

    input_output_path = os.path.dirname(__file__)
    os.environ["LANDBOSSE_INPUT_DIR"] = input_output_path
    os.environ["LANDBOSSE_OUTPUT_DIR"] = input_output_path

    extended_project_list_before_parameter_modifications = read_data(input_dict)
    xlsx_reader = XlsxReader()

    for _, project_parameters in \
            extended_project_list_before_parameter_modifications.iterrows():
        # If project_parameters['Project ID with serial'] is null, that means
        # there are no parametric modifications to the project data dataframes.
        # Hence, just the plain Project ID without a serial number should be
        # used.
        if pd.isnull(project_parameters['Project ID with serial']):
            project_id_with_serial = project_parameters['Project ID']
        else:
            project_id_with_serial = project_parameters['Project ID with serial']
            # Read the project data sheets.

        project_data_basename = project_parameters['Project data file']

        project_data_sheets = \
            XlsxDataframeCache.read_all_sheets_from_xlsx(project_data_basename)

        # make sure you call create_master_input_dictionary() as soon as
        # labor_cost_multiplier's value is changed.
        # The weather dataframe key in input_dict is 'user_weather_DF':
        master_input_dict = xlsx_reader.create_master_input_dictionary(project_data_sheets,
                                                                       project_parameters)
        master_input_dict['error'] = dict()

    output_dict = dict()
    default_date_time = list()
    # <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
    # override default inputs with user modified inputs (on the SAM UI):
    if 'project_id' in input_dict:
        project_id_with_serial = input_dict['project_id']

    try:
        for key, value in input_dict.items():
            if (type(value) == int or type(value) == float) and value < 0:
                raise NegativeInputError
    except NegativeInputError:
        master_input_dict['error']['NegativeInputError'] = \
            'User entered a negative value for ' + key + '. This is an invalid entry'

    # Loop to collect user provided inputs that will override defaults:

    for key, _ in input_dict.items():

        if key == 'weather_file_path':  # weather_file_path is handled later in
            # this codebase. Skip storing it for now.
            pass

        elif key in master_input_dict:
            master_input_dict[key] = input_dict[key]

        elif key == 'path_to_project_list' or key == 'name_of_project_list':
            pass

        # Allows user to override default development cost:
        elif key == 'development_labor_cost_usd':
            master_input_dict[key] = input_dict['development_labor_cost_usd']

        # New functionality that lets a user run model with a user defined
        # grid interconnection rating:
        elif key == 'grid_system_size_MW':
            master_input_dict['grid_system_size_MW'] = input_dict['grid_system_size_MW']

        elif key == 'substation_rating_MW':
            master_input_dict['substation_rating_MW'] = input_dict['substation_rating_MW']

        else:
            exit(1)

    # update master input dict based on new labor cost multiplier:
    labor_cost_multiplier = master_input_dict['labor_cost_multiplier']
    xlsx_reader.apply_labor_multiplier_to_project_data_dict(project_data_sheets,
                                                            labor_cost_multiplier)

    # Ensuring number of turbines is > 10:
    try:
        if 'num_turbines' in input_dict:
            if input_dict['num_turbines'] > 0:
                master_input_dict['num_turbines'] = input_dict['num_turbines']
            else:
                raise TurbineNumberError
    except TurbineNumberError:
        master_input_dict['error']['TurbineNumberError'] = \
            'User selected less than 1 turbine. LandBOSSE currently ' \
            'provides BOS estimates for 1 or greater number of turbines.'

    # Ensuring user is runnning landbosse for turbine in range of [1- 8] MW:
    try:
        if 'turbine_rating_MW' in input_dict:
            if input_dict['turbine_rating_MW'] > 8:
                raise LargeTurbineSizeError
            elif input_dict['turbine_rating_MW'] < 1:
                raise SmallTurbineSizeError
            else:
                master_input_dict['turbine_rating_MW'] = input_dict['turbine_rating_MW']
    except LargeTurbineSizeError:
        master_input_dict['error']['LargeTurbineSizeError'] = \
                        'User selected turbine of rating greater than 8 MW. ' \
                        'LandBOSSE provides reasonable results for turbines ' \
                        'in range of 1-8 MW rating for now.'

    except SmallTurbineSizeError:
        master_input_dict['error']['SmallTurbineSizeError'] = \
                        'User selected turbine of rating lesser than 1 MW. ' \
                        'LandBOSSE provides reasonable results for turbines ' \
                        'in range of 1-8 MW rating for now.'

    # Weather file passed by SAM does NOT have a 'Date UTC' column. So it needs
    # to be manually added to prevent code from breaking. Needs to have 8760
    # rows worth of date_time to stay consistent with shape of SAM weather data
    # passed.
    start_date = datetime(2011, 12, 31, 00, 00)
    end_date = datetime(2012, 12, 30, 00, 00)
    default_date_time = []
    for single_date in daterange(start_date, end_date):
        default_date_time.append(single_date.strftime("%Y-%m-%d %H:%M"))

    # make sure weather data passed by SAM is hourly.
    try:
        if 'project_id' in input_dict:
            pass

        else:
            if 'weather_file_path' in input_dict:  # if user provides weather file, use it
                weather_window_input_df = read_weather_data(input_dict['weather_file_path'])

            elif 'user_weather_DF' in input_dict:  # if user provides weather DF, use it
                weather_window_input_df = input_dict['user_weather_DF']

            else:   # if user doesn't provide weather data, use default file
                weather_file_path = input_output_path + \
                                    '/project_data/az_rolling.srw'
                weather_window_input_df = read_weather_data(weather_file_path)


            weather_window_input_df = weather_window_input_df.reset_index(
                                                                    drop=True)
            weather_window_input_df.insert(loc=0,
                                           column='time',
                                           value=default_date_time
                                           )
            column_names = weather_window_input_df.columns
            renamed_columns = {
                column_names[0]: 'Date UTC',
                column_names[1]: 'Temp C',
                column_names[2]: 'Pressure atm',
                column_names[3]: 'Direction deg',
                column_names[4]: 'Speed m per s'
            }
            weather_data = weather_window_input_df.rename(columns=renamed_columns)
            weather_data = weather_data.reset_index(drop=True)
            weather_data = weather_data[renamed_columns.values()]
            master_input_dict['weather_window'] = read_weather_window(weather_data)
    except Exception as error:  # exception handling for landbosse_api
        master_input_dict['error']['Weather_Data'] = error

    # <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
    # If user provides 'project_id' in input_dict, it is because they want to use the
    # default detailed data input file either shipped w/ LandBOSSE, or provided by the
    # user. In this case, skip re-scaling of turbine components.
    #
    # ELSE, if user doesn't provide 'project_id', use the scaling relations to refactor
    # Components DF based on user inputs.

    if 'project_id' in input_dict:

        pass

    # Else, use the scaling relations to refactor Components DF based on user inputs:
    else:
        # nacelle
        nacelle_mass_ton = nacelle_mass(master_input_dict['turbine_rating_MW'])
        master_input_dict['component_data'] = edit_nacelle_info(
                                            master_input_dict['component_data'],
                                            nacelle_mass_ton,
                                            master_input_dict['hub_height_meters']
                                            )

        master_input_dict['component_data'] = master_input_dict[
                                        'component_data'].reset_index(drop=True)

        # hub
        hub_mass_ton = hub_mass(master_input_dict['turbine_rating_MW'])
        master_input_dict['component_data'] = edit_hub_info(
                                            master_input_dict['component_data'],
                                            hub_mass_ton,
                                            master_input_dict['hub_height_meters']
                                            )

        master_input_dict['component_data'] = master_input_dict[
                                        'component_data'].reset_index(drop=True)

        # combined 3 blades total mass in tons:
        blade_mass = blade_mass_ton(master_input_dict['rotor_diameter_m'])
        master_input_dict['component_data'] = edit_blade_info(
                                            master_input_dict['component_data'],
                                            blade_mass,
                                            master_input_dict['hub_height_meters'],
                                            master_input_dict['rotor_diameter_m']
                                            )

        master_input_dict['component_data'] = master_input_dict[
                                        'component_data'].reset_index(drop=True)

        # tower
        tower_mass_tonne = tower_mass(master_input_dict['turbine_rating_MW'],
                                      master_input_dict['hub_height_meters']
                                      )

        num_tower_sections, tower_section_height_m = tower_specs(
                                            master_input_dict['hub_height_meters'],
                                            tower_mass_tonne
                                            )

        master_input_dict['component_data'] = edit_tower_sections(
                                            master_input_dict['component_data'],
                                            num_tower_sections,
                                            tower_mass_tonne,
                                            tower_section_height_m
                                            )

        master_input_dict['component_data'] = master_input_dict[
                                        'component_data'].reset_index(drop=True)
    # <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><

    # Ensure plant size is getting updated based on user provided num turbines and
    # turbine rating
    master_input_dict['project_size_megawatts'] = master_input_dict['num_turbines'] * \
                                                  master_input_dict['turbine_rating_MW']

    master_input_dict['plant_capacity_MW'] = master_input_dict['num_turbines'] * \
                                                  master_input_dict['turbine_rating_MW']

    # Manager class (1) manages the distribution of inout data for all modules
    # and (2) executes landbosse
    mc = Manager(input_dict=master_input_dict, output_dict=output_dict)
    mc.execute_landbosse(project_id_with_serial)

    # results dictionary that gets returned by this function:
    results = dict()
    results['errors'] = []
    if master_input_dict['error']:
        for key, value in master_input_dict['error'].items():
            msg = "Error in " + key + ": " + str(value)
            results['errors'].append(msg)
    else:
        # if project runs successfully, return a dictionary with results that are 3
        # layers of results deep (but 1-D). The layers of results here refers to:
        # 1) Total BOS cost
        # 2) Total module cost (that is, foundation cost, collection cost etc.)
        # 3) Breakdown of each module cost into:
        #   a) Labor
        #   b) Equipment rental
        #   c) Materials
        #   d) Mobilization
        #   e) Other [only in some modules]
        #   f) Fuel [only in some modules]

        results['total_bos_cost'] = \
            output_dict['project_value_usd']

        # management cost module results:
        results['total_management_cost'] = \
            output_dict['total_management_cost']

        results['insurance_usd'] = \
            output_dict['insurance_usd']

        results['construction_permitting_usd'] = \
            output_dict['construction_permitting_usd']

        results['project_management_usd'] = \
            output_dict['project_management_usd']

        results['bonding_usd'] = \
            output_dict['bonding_usd']

        results['markup_contingency_usd'] = \
            output_dict['markup_contingency_usd']

        results['engineering_usd'] = \
            output_dict['engineering_usd']

        results['site_facility_usd'] = \
            output_dict['site_facility_usd']

        results['total_management_cost'] = \
            output_dict['total_management_cost']

        # development cost module results:
        results['total_development_cost'] = \
            output_dict['summed_development_cost']

        if 'Development labor cost USD' in output_dict:
            results['development_equipment_rental_usd'] = \
                master_input_dict['development_labor_cost_usd']

        results['development_labor_usd'] = 0
        results['development_material_usd'] = 0
        results['development_mobilization_usd'] = 0

        # site prep cost module results:
        results['total_sitepreparation_cost'] = \
            output_dict['summed_sitepreparation_cost']

        results['sitepreparation_equipment_rental_usd'] = \
            output_dict['sitepreparation_equipment_rental_usd']

        results['sitepreparation_labor_usd'] = \
            output_dict['sitepreparation_labor_usd']

        results['sitepreparation_material_usd'] = \
            output_dict['sitepreparation_material_usd']

        results['sitepreparation_mobilization_usd'] = \
            output_dict['sitepreparation_mobilization_usd']

        results['sitepreparation_other_usd'] = \
            output_dict['sitepreparation_other_usd']

        # foundation cost module results:
        results['total_foundation_cost'] = \
            output_dict['summed_foundation_cost']

        results['foundation_equipment_rental_usd'] = \
            output_dict['foundation_equipment_rental_usd']

        results['foundation_labor_usd'] = \
            output_dict['foundation_labor_usd']

        results['foundation_material_usd'] = \
            output_dict['foundation_material_usd']

        results['foundation_mobilization_usd'] = \
            output_dict['foundation_mobilization_usd']

        # erection cost module results:
        results['total_erection_cost'] = \
            output_dict['total_cost_summed_erection']

        results['erection_equipment_rental_usd'] = \
            output_dict['erection_equipment_rental_usd']

        results['erection_labor_usd'] = \
            output_dict['erection_labor_usd']

        results['erection_material_usd'] = \
            output_dict['erection_material_usd']

        results['erection_other_usd'] = \
            output_dict['erection_other_usd']

        results['erection_mobilization_usd'] = \
            output_dict['erection_mobilization_usd']

        results['erection_fuel_usd'] = \
            output_dict['erection_fuel_usd']

        # grid connection cost module results:
        results['total_gridconnection_cost'] = \
            output_dict['trans_dist_usd']

        # collection cost module results:
        results['total_collection_cost'] = \
            output_dict['summed_collection_cost']

        results['collection_equipment_rental_usd'] = \
            output_dict['collection_equipment_rental_usd']

        results['collection_labor_usd'] = \
            output_dict['collection_labor_usd']

        results['collection_material_usd'] = \
            output_dict['collection_material_usd']

        results['collection_mobilization_usd'] = \
            output_dict['collection_mobilization_usd']

        # substation cost module results:
        results['total_substation_cost'] = \
            output_dict['summed_substation_cost']

    return results


# This method reads in the two input Excel files (project_list; project_1)
# and stores them as data frames. This method is called internally in
# run_landbosse(), where the data read in is converted to a master input
# dictionary.
def read_data(modified_path_dict):
    if 'path_to_project_list' in modified_path_dict:
        path_to_project_list = modified_path_dict['path_to_project_list']
    else:
        path_to_project_list = os.path.dirname(__file__)

    if 'name_of_project_list' in modified_path_dict:
        project_list = modified_path_dict['name_of_project_list']
    else:
        project_list = 'project_list'

    sheets = XlsxDataframeCache.read_all_sheets_from_xlsx(project_list,
                                                          path_to_project_list)

    # If there is one sheet, make an empty dataframe as a placeholder.
    if len(sheets.values()) == 1:
        first_sheet = list(sheets.values())[0]
        project_list = first_sheet
        parametric_list = pd.DataFrame()

    # If the parametric and project lists exist, read them
    elif 'Parametric list' in sheets.keys() \
            and 'Project list' in sheets.keys():

        project_list = sheets['Project list']
        parametric_list = sheets['Parametric list']

    # Otherwise, raise an exception
    else:
        raise KeyError(
            "Project list needs to have a single sheet or sheets named "
            "'Project list' and 'Parametric list'."
        )

    # Instantiate and XlsxReader to assemble master input dictionary
    xlsx_reader = XlsxReader()

    # Join in the parametric variable modifications
    parametric_value_list = xlsx_reader.create_parametric_value_list(parametric_list)
    extended_project_list = xlsx_reader.\
        outer_join_projects_to_parametric_values(project_list,
                                                 parametric_value_list)

    return extended_project_list


def read_weather_data(file_path):
    weather_data = pd.read_csv(file_path,
                               sep=",",
                               header=None,
                               skiprows=5,
                               usecols=[0, 1, 2, 3, 4]
                               )
    return weather_data


# Weather file passed by SAM does NOT have a 'Date UTC' column. So it needs to
# be manually added to prevent code from breaking. Needs to have 8765 rows
# worth of date_time to stay consistent with shape of SAM weather data passed.


def daterange(start_date, end_date):
    delta = timedelta(hours=1)
    while start_date < end_date:
        yield start_date
        start_date += delta


class Error(Exception):
    """
        Base class for other exceptions
    """
    pass


class SmallTurbineSizeError(Error):
    """
        Raised when user selects a turbine of rating less than 1 MW since
        LandBOSSE provides reasonable results for turbines greater than
        1 MW rating for now.
    """
    pass


class LargeTurbineSizeError(Error):
    """
            Raised when user selects a turbine of rating less than 8 MW
            since LandBOSSE provides reasonable results for turbines greater
            than 1 MW rating for now.
    """
    pass


class TurbineNumberError(Error):
    """
        Raised when number of turbines is less than 10; since LandBOSSE API
        does not currently handle BOS calculations for < 10 turbines.
    """
    pass


class NegativeInputError(Error):
    """
        User entered a negative input. This is an invalid entry.
    """
    pass


# <><><><><><><><> EXAMPLE OF RUNNING THIS LANDBOSSE API <><><><><><><><><><><>

# Default inputs on the SAM UI. Commented out since SAM will pass these values
# down to LandBOSSE.
# TODO: Un-comment these out if running this script directly.
# input_dict = dict()
# input_dict['interconnect_voltage_kV'] = 137
# input_dict['distance_to_interconnect_mi'] = 0.31
# input_dict['num_turbines'] = 1
# input_dict['project_id'] = 'ge15_public_dist'
# input_dict['turbine_spacing_rotor_diameters'] = 4
# input_dict['row_spacing_rotor_diameters'] = 10
# input_dict['turbine_rating_MW'] = 2.5
# input_dict['rotor_diameter_m'] = 77
# input_dict['hub_height_meters'] = 80
# input_dict['wind_shear_exponent'] = 0.20
# input_dict['depth'] = 2.36  # Foundation depth in m
# input_dict['rated_thrust_N'] = 589000
# input_dict['labor_cost_multiplier'] = 1
# input_dict['gust_velocity_m_per_s'] = 59.50
# input_dict['path_to_project_list'] = '/Users/<username>/Desktop/...'
# input_dict['name_of_project_list'] = 'project_list_alternative'


# (Optional) Provide absolute file path of wind weather file (.txt, .srw, or
# .csv). Wind data used here follows the wind toolkit (WTK) formatted data.
# If you do not provide a path to a weather file, this API will use the
# default weather file shipped with LandBOSSE.

# Alternatively, provide a [8760 x 4] dataframe with WTK formatted data:
# [Temp (C), Pressure (atm), Wind Direction (deg), Wind Speed(m/s)]

# Example of how to provide path to weather file:
# input_dict['weather_file_path'] = '/Users/<username>/Desktop/az_rolling.srw'

# Example of how to provide weather input dataframe:
# input_dict['user_weather_DF'] = pd.DataFrame(np.column_stack((15*np.ones((8760, 1)),
#                                                               np.ones((8760, 1)),
#                                                               180*np.ones((8760, 1)),
#                                                               9*np.ones((8760, 1)))))
#
# BOS_results = run_landbosse(input_dict)
# print(BOS_results)
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><



from collections import OrderedDict
import os

import pandas as pd

from ..model import Manager
from .XlsxFileOperations import XlsxFileOperations
from .XlsxReader import XlsxReader
from .XlsxManagerRunner import XlsxManagerRunner
from .XlsxDataframeCache import XlsxDataframeCache
from .XlsxGenerator import XlsxGenerator


class XlsxSerialManagerRunner(XlsxManagerRunner):
    """
    This subclass implementation of XlsxManagerRunner runs all projects
    in a serial loop.
    """

    def run_from_project_list_xlsx(self, projects_xlsx, enable_cost_and_scaling_modifications=False):
        """
        This function runs all the scenarios in the projects_xlsx file. It creates
        the OrderedDict that holds the results of all the runs. See the return
        section below for more details on what the OrderedDict contains.

        This is a concrete implementation of the super class method.

        Parameters
        ----------
        projects_xlsx : str
            A path name (preferably created with os.path.join()) specific to the
            operating system that is the main input .xlsx file that controls
            running of all the projects. Crucially, this file contains names of
            other. It is recommended that all input file be kept in the same
            input directory. Each line of projects_xlsx becomes a project_series.

        enable_cost_and_scaling_modifications : bool
            If True, this method modifies each row of the project list AFTER it has been
            modified by the parameters for to scale certain input values based
            on what has been parametrically modified.

        Returns
        -------
        OrderedDict, list, list, list
            First element of tuple is an ordered dict that is the result of
            all the runs. Each key is the name of a project and each value
            is the output dictionary of that project. The second element
            is the list of rows for the csv. The third element is the list
            of costs for the spreadsheets. The fourth element is the same as
            module_type_operation_lists, but every row has all the inputs
            on each row.
        """
        # Load the project list
        extended_project_list_before_parameter_modifications = self.read_project_and_parametric_list_from_xlsx()
        print('>>> Project and parametric lists loaded')

        # For file operations
        file_ops = XlsxFileOperations()

        # Get the output dictionary ready
        runs_dict = OrderedDict()

        # Instantiate and XlsxReader to assemble master input dictionary
        xlsx_reader = XlsxReader()

        # Get a list ready to hold the project parameters after they have been modified
        # After all rows have been added to this list (each row is a series) then the
        # whole list will be transformed into a dataframe.
        #
        # See notes at https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.append.html
        # for why this is more performant than appending to a dataframe.
        extended_project_list_after_parameter_modifications = []

        # Loop over every project
        for _, project_parameters in extended_project_list_before_parameter_modifications.iterrows():

            # If project_parameters['Project ID with serial'] is null, that means there are no
            # parametric modifications to the project data dataframes. Hence,
            # just the plain Project ID without a serial number should be used.
            if pd.isnull(project_parameters['Project ID with serial']):
                project_id_with_serial = project_parameters['Project ID']
            else:
                project_id_with_serial = project_parameters['Project ID with serial']

            project_data_basename = project_parameters['Project data file']

            # Input path for unmodified project input data.
            project_data_xlsx = os.path.join(file_ops.landbosse_input_dir(), 'project_data', f'{project_data_basename}.xlsx')

            # Log each project
            print(f'<><><><><><><><><><><><><><><><><><> {project_id_with_serial} <><><><><><><><><><><><><><><><><><>')
            print('>>> project_id: {}'.format(project_id_with_serial))
            print('>>> Project data: {}'.format(project_data_xlsx))

            # Read the project data sheets.
            project_data_sheets = XlsxDataframeCache.read_all_sheets_from_xlsx(project_data_basename)

            # Transform the dataframes so that they have the right values for
            # the parametric variables.
            xlsx_reader.modify_project_data_and_project_list(project_data_sheets, project_parameters)

            # Apply cost and scaling modifications if needed.
            if enable_cost_and_scaling_modifications:
                xlsx_reader.apply_cost_and_scaling_modifications_to_project_parameters(project_parameters)

            # Append the modified project parameters
            extended_project_list_after_parameter_modifications.append(project_parameters)

            # Write all project_data sheets
            parametric_project_data_path = \
                os.path.join(file_ops.parametric_project_data_output_path(), f'{project_id_with_serial}_project_data.xlsx')
            XlsxGenerator.write_project_data(project_data_sheets, parametric_project_data_path)

            # Create the master input dictionary.
            master_input_dict = xlsx_reader.create_master_input_dictionary(project_data_sheets, project_parameters)

            # Now run the manager and accumulate its result into the runs_dict
            output_dict = dict()
            mc = Manager(input_dict=master_input_dict, output_dict=output_dict)
            mc.execute_landbosse(project_name=project_id_with_serial)
            output_dict['project_series'] = project_parameters
            runs_dict[project_id_with_serial] = output_dict

        final_result = dict()
        final_result['details_list'] = self.extract_details_lists(runs_dict)
        final_result['module_type_operation_list'] = self.extract_module_type_operation_lists(runs_dict)
        final_result['extended_project_list'] = pd.DataFrame(extended_project_list_after_parameter_modifications)

        # Return the runs for all the projects.
        return final_result



import math

class CostModule:
    """
    This is a super class for all other cost modules to import
    that provides shared methods for outputs from results and
    mobilization cost calculations.
    """

    def mobilization_cost_multiplier(self, turbine_rating):
        """
        Calculates a mobilization cost term as a function of
        turbine rating.

        Parameters
        ----------
        turbine_rating : float
            Turbine rating in megawatts

        Returns
        -------
        float
            The mobilization cost multiplier as a function of turbine rating.
        """

        mobilization_cost_multiplier = (36.892 * math.exp(-5e-04 * (turbine_rating * 1000))) / 100
        return mobilization_cost_multiplier

    def outputs_for_costs_by_module_type_operation(self,
                                                   *,
                                                   input_df,
                                                   project_id,
                                                   total_or_turbine):
        """
        This takes a dataframe and turns it into a list of dictionaries
        suitable for output to a cost tab in a spreadsheet.

        Outputs dictionaries that are rows, with each row being a dict that
        has costs broken down by and module id, operation id, type of cost,
        cost, and per turbine or total. Each of those values are stored in
        their own column.

        It maps each row of this dataframe into a dictionary with keys
        of those names. It accumulates these dictionaries into a list.

        It must be called with keyword arguments.

        Parameters
        ----------
        input_df : pd.DataFrame
           The input dataframe that has the columns listed above.

        project_id : str
            The id of the project (it is a string, not an integer) to
            place in each row.

        total_or_turbine : bool
            True if the cost is total for the project. False if it
            is per turbine.

        Returns
        -------
        list
            List of dicts, with each dict representing a row for
            the output.
        """
        result = []
        # module = type(self).__name__
        module = 'CollectionCost' if (type(self).__name__ == 'ArraySystem') else type(self).__name__
        turbine_rating_MW = self.input_dict['turbine_rating_MW']
        num_turbines = self.input_dict['num_turbines']
        rotor_diameter_m = self.input_dict['rotor_diameter_m']
        project_size_kw = num_turbines * turbine_rating_MW * 1000

        for _, row in input_df.iterrows():
            _dict = dict()
            row = row.to_dict()
            _dict['operation_id'] = row['Phase of construction']
            _dict['type_of_cost'] = row['Type of cost']
            _dict['raw_cost'] = row['Cost USD']
            result.append(_dict)

        for _dict in result:
            _dict['turbine_rating_MW'] = turbine_rating_MW
            _dict['num_turbines'] = num_turbines
            _dict['rotor_diameter_m'] = rotor_diameter_m
            _dict['project_id_with_serial'] = self.project_name
            _dict['module'] = module

            if total_or_turbine:  # If raw_cost is the total cost
                _dict['raw_cost_total_or_per_turbine'] = 'total'
                _dict['cost_per_turbine'] = _dict['raw_cost'] / num_turbines
                _dict['cost_per_project'] = _dict['raw_cost']
                _dict['usd_per_kw_per_project'] = _dict['raw_cost'] / project_size_kw
            else:                 # If raw_cost is per turbine
                _dict['raw_cost_total_or_per_turbine'] = 'turbine'
                _dict['cost_per_turbine'] = _dict['raw_cost']
                _dict['cost_per_project'] = _dict['raw_cost'] * num_turbines
                _dict['usd_per_kw_per_project'] = _dict['cost_per_project'] / project_size_kw

        return result



import os
from concurrent import futures

import pandas as pd

from ..model import Manager
from .XlsxFileOperations import XlsxFileOperations
from .XlsxReader import XlsxReader
from .XlsxManagerRunner import XlsxManagerRunner
from .XlsxDataframeCache import XlsxDataframeCache
from .XlsxGenerator import XlsxGenerator


class XlsxParallelManagerRunner(XlsxManagerRunner):
    """
    This subclass implementation of XlsxManagerRunner runs all projects
    with a ProcessPoolExecutor.
    """

    def run_from_project_list_xlsx(self, projects_xlsx, enable_cost_and_scaling_modifications=False):
        """
        This function runs all the scenarios in the projects_xlsx file. It creates
        the OrderedDict that holds the results of all the runs. See the return
        section below for more details on what the OrderedDict contains.

        This is a concrete implementation of the super class method.

        Parameters
        ----------
        projects_xlsx : str
            A path name (preferably created with os.path.join()) specific to the
            operating system that is the main input .xlsx file that controls
            running of all the projects. Crucially, this file contains names of
            other. It is recommended that all input file be kept in the same
            input directory.

        enable_cost_and_scaling_modifications : bool
            If True, this method modifies each row of the project list AFTER it has been
            modified by the parameters for to scale certain input values based
            on what has been parametrically modified.

        Returns
        -------
        dict, list, module_type_operation_lists
            First element of tuple is a dict that is the result of
            all the runs. Each key is the name of a project and each value
            is the output dictionary of that project. The second element
            is the list of rows for the csv. The third element is the list
            of costs for the spreadsheets.
        """
        # Load the project list
        print('Calculating parametric values')
        extended_project_list_before_parameter_modifications = self.read_project_and_parametric_list_from_xlsx()

        # Prepare the file operations
        file_ops = XlsxFileOperations()

        # Instantiate an XlsxReader to handle the parametrics and master input
        # dictionaries
        xlsx_reader = XlsxReader()

        # Get a list ready to hold the project parameters after they have been modified
        # After all rows have been added to this list (each row is a series) then the
        # whole list will be transformed into a dataframe.
        #
        # See notes at https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.append.html
        # for why this is more performant than appending to a dataframe.
        extended_project_list_after_parameter_modifications = []

        # Prep all task for the executor
        all_tasks = []
        print(f'Found {len(extended_project_list_before_parameter_modifications)} projects for execution')
        for _, project_parameters in extended_project_list_before_parameter_modifications.iterrows():

            # If project_parameters['Project ID with serial'] is null, that means there are no
            # parametric modifications to the project data dataframes. Hence,
            # just the plain Project ID without a serial number should be used.
            if pd.isnull(project_parameters['Project ID with serial']):
                project_id_with_serial = project_parameters['Project ID']
            else:
                project_id_with_serial = project_parameters['Project ID with serial']

            print(f'Preparing {project_id_with_serial}')

            project_data_basename = project_parameters['Project data file']
            task = dict()

            task['project_data_sheets'] = XlsxDataframeCache.read_all_sheets_from_xlsx(project_data_basename)

            # Transform the dataframes so that they have the right values for
            # the parametric variables.
            xlsx_reader.modify_project_data_and_project_list(task['project_data_sheets'], project_parameters)

            # Apply cost and scaling modifications if needed.
            if enable_cost_and_scaling_modifications:
                xlsx_reader.apply_cost_and_scaling_modifications_to_project_parameters(project_parameters)

            # Append the modified project parameters
            extended_project_list_after_parameter_modifications.append(project_parameters)

            # Write all project_data sheets
            parametric_project_data_path = \
                os.path.join(file_ops.parametric_project_data_output_path(), f'{project_id_with_serial}_project_data.xlsx')
            XlsxGenerator.write_project_data(task['project_data_sheets'], parametric_project_data_path)

            task['project_data_basename'] = project_data_basename
            task['project_id_with_serial'] = project_id_with_serial
            task['project_series'] = project_parameters
            all_tasks.append(task)

        # Execute every project
        with futures.ProcessPoolExecutor() as executor:
            executor_result = executor.map(run_single_project, all_tasks)

        # Get the output dictionary ready
        runs_dict = {project_id_with_serial: result for project_id_with_serial, result in executor_result}

        # Assemble the dictionary with content for the details, details with inputs,
        #  cost_by_module_type_operation and cost_by_module_type_operation_with_input tabs
        final_result = dict()
        final_result['details_list'] = self.extract_details_lists(runs_dict)
        final_result['module_type_operation_list'] = self.extract_module_type_operation_lists(runs_dict)
        final_result['extended_project_list'] = pd.DataFrame(extended_project_list_after_parameter_modifications)

        # Return the runs for all the scenarios.
        return final_result


"""
The following function is deliberately defined outside of the class.
This makes it easier to think about it being a pure function for
parallel processes.
"""


def run_single_project(task_dict):
    """
    The dictionary project_definition_dict contains the following keys.

    For each process another logger is created, so that each process does
    not attempt to use the same logger.

    input_xlsx : str
        The filename for the input .xlsx that has all the dataframes
        for the for ErectionCost and FoundationCost

    project_series : pd.Series
        The series that has the non-dataframe values for each project,
        including the project name.

    project_id : str
        The string that is the name of the project.

    Basically, the map operation goes like this:

    task_dict -> master_input_dict -> master_output_dict

    Wrapped in a functional executor, this maps projects into their
    output dictionaries.

    Parameters
    ----------
    task_dict : dict
        The configuration of the task.

    Returns
    -------
    tuple : (str, dict)
        The str is the project_id. The dict is the resulting output
        dictionary.
    """
    project_data_basename = task_dict['project_data_basename']
    project_series = task_dict['project_series']
    project_id_with_serial = task_dict['project_id_with_serial']
    project_data_sheets = task_dict['project_data_sheets']

    # Log each project. Use print because it works better for multiple processes.
    print(f'Start {project_id_with_serial}, project data in {project_data_basename}')

    # Read the Excel
    xlsx_reader = XlsxReader()
    master_input_dict = xlsx_reader.create_master_input_dictionary(project_data_sheets, project_series)

    # Now run the manager and accumulate its result into the runs_dict
    output_dict = dict()
    output_dict['project_series'] = project_series
    mc = Manager(input_dict=master_input_dict, output_dict=output_dict)
    mc.execute_landbosse(project_name=project_id_with_serial)

    print(f'End {project_id_with_serial}')

    return project_id_with_serial, output_dict



import os
import setuptools

name = 'landbosse'
version = '2.3.0'

with open('README.md', 'r') as fh:
    long_description = fh.read()

PACKAGE_PATH = os.path.abspath(os.path.join(__file__, os.pardir))

setuptools.setup(
    url='https://github.com/WISDEM/LandBOSSE',
    name=name,
    version=version,
    author='NREL',
    author_email='Parangat.Bhaskar@nrel.gov',
    include_package_data=True,
    python_requires='>=3.7',
    description='LandBOSSE',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['landbosse', 'landbosse.model','landbosse.excelio','landbosse.tests', 'landbosse.landbosse_api'],
    # packages=setuptools.find_packages(PACKAGE_PATH, "test"),
    test_suite='nose.collector',
    tests_require=['nose'],
    install_requires=[
        'pandas==1.0.3',
        'numpy==1.17.2',
        'sympy==1.4',
        'scipy==1.3.1',
        'xlsxwriter==1.2.1',
        'xlrd==1.2.0',
        'pytest==5.3.5'
        # 'pscyopg2-binary==2.8.4',
        # 'sqlalchemy==1.3.13'
    ],

    command_options={
            'build_sphinx': {
                'project': ('setup.py', name),
                'version': ('setup.py', version),
                # 'release': ('setup.py', release),  # Not yet needed
                'source_dir': ('setup.py', 'docs')}}
)
